import pytest
from .utils import helper

from hera.data import Token
from hera.op import name_to_class
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def branch_helper(vm, branchname):
    op = name_to_class[branchname](Token.Sym("l"))
    return op.should(vm)


def test_exec_branch_with_register_branch(vm):
    vm.registers[7] = 42
    vm.flag_zero = True

    helper(vm, "BZ(R7)")

    assert vm.pc == 42
    assert vm.flag_zero


def test_exec_branch_with_relative_branch(vm):
    vm.pc = 20
    vm.flag_carry = False

    helper(vm, "BNCR(22)")

    assert vm.pc == 42
    assert not vm.flag_carry


def test_should_BL_on_sign(vm):
    vm.flag_sign = True
    assert branch_helper(vm, "BL")


def test_should_BL_on_overflow(vm):
    vm.flag_overflow = True
    assert branch_helper(vm, "BL")


def test_should_not_BL_on_sign_and_overflow(vm):
    vm.flag_sign = True
    vm.flag_overflow = True
    assert not branch_helper(vm, "BL")


def test_should_not_BL_on_neither_sign_nor_overflow(vm):
    vm.flag_sign = False
    vm.flag_overflow = False
    assert not branch_helper(vm, "BL")


def test_should_BGE_on_no_flags(vm):
    assert branch_helper(vm, "BGE")


def test_should_BGE_on_sign_and_overflow(vm):
    vm.flag_overflow = True
    vm.flag_sign = True
    assert branch_helper(vm, "BGE")


def test_should_not_BGE_on_sign(vm):
    vm.flag_sign = True
    assert not branch_helper(vm, "BGE")


def test_should_not_BGE_on_overflow(vm):
    vm.flag_overflow = True
    assert not branch_helper(vm, "BGE")


def test_should_BGE_on_zero(vm):
    vm.flag_zero = True
    assert branch_helper(vm, "BGE")


def test_should_BLE_on_sign(vm):
    vm.flag_sign = True
    assert branch_helper(vm, "BLE")


def test_should_BLE_on_overflow(vm):
    vm.flag_overflow = True
    assert branch_helper(vm, "BLE")


def test_should_BLE_on_zero(vm):
    vm.flag_zero = True
    assert branch_helper(vm, "BLE")


def test_should_BLE_on_overflow_and_zero(vm):
    vm.flag_overflow = True
    vm.flag_zero = True
    assert branch_helper(vm, "BLE")


def test_should_not_BLE_on_no_flags(vm):
    assert not branch_helper(vm, "BLE")


def test_should_not_BLE_on_sign_and_overflow(vm):
    vm.flag_sign = True
    vm.flag_overflow = True
    assert not branch_helper(vm, "BLE")


def test_should_BG_on_no_flags(vm):
    assert branch_helper(vm, "BG")


def test_should_BG_on_sign_and_overflow(vm):
    vm.flag_overflow = True
    vm.flag_sign = True
    assert branch_helper(vm, "BG")


def test_should_not_BG_on_sign(vm):
    vm.flag_sign = True
    assert not branch_helper(vm, "BG")


def test_should_not_BG_on_overflow(vm):
    vm.flag_overflow = True
    assert not branch_helper(vm, "BG")


def test_should_not_BG_on_zero(vm):
    vm.flag_zero = True
    assert not branch_helper(vm, "BG")


def test_should_BULE_on_not_carry(vm):
    vm.flag_carry = False
    assert branch_helper(vm, "BULE")


def test_should_not_BULE_on_overflow(vm):
    vm.flag_carry = True
    vm.flag_overflow = True
    assert not branch_helper(vm, "BULE")


def test_should_BULE_on_zero(vm):
    vm.flag_carry = True
    vm.flag_zero = True
    assert branch_helper(vm, "BULE")


def test_should_not_BULE_on_sign_and_carry(vm):
    vm.flag_carry = True
    vm.flag_sign = True
    assert not branch_helper(vm, "BULE")


def test_should_BUG_on_carry_and_not_zero(vm):
    vm.flag_carry = True
    vm.flag_zero = False
    assert branch_helper(vm, "BUG")


def test_should_not_BUG_on_carry_and_zero(vm):
    vm.flag_carry = True
    vm.flag_zero = True
    assert not branch_helper(vm, "BUG")


def test_should_not_BUG_on_not_carry_and_not_zero(vm):
    vm.flag_carry = False
    vm.flag_zero = False
    assert not branch_helper(vm, "BUG")


def test_should_not_BUG_on_sign(vm):
    vm.flag_sign = True
    assert not branch_helper(vm, "BUG")


def test_should_BZ_on_zero(vm):
    vm.flag_zero = True
    assert branch_helper(vm, "BZ")


def test_should_not_BZ_on_not_zero(vm):
    vm.flag_zero = False
    assert not branch_helper(vm, "BZ")


def test_should_BC_on_carry(vm):
    vm.flag_carry = True
    assert branch_helper(vm, "BC")


def test_should_not_BC_on_not_carry(vm):
    vm.flag_carry = False
    assert not branch_helper(vm, "BC")


def test_should_BS_on_sign(vm):
    vm.flag_sign = True
    assert branch_helper(vm, "BS")


def test_should_not_BS_on_not_sign(vm):
    vm.flag_sign = False
    assert not branch_helper(vm, "BS")


def test_should_BV_on_overflow(vm):
    vm.flag_overflow = True
    assert branch_helper(vm, "BV")


def test_should_not_BV_on_not_overflow(vm):
    vm.flag_overflow = False
    assert not branch_helper(vm, "BV")


def test_CALL_changes_pc(vm):
    vm.pc = 100
    vm.registers[13] = 40

    helper(vm, "CALL(R12, R13)")

    assert vm.pc == 40


def test_CALL_updates_second_register(vm):
    vm.pc = 100
    vm.registers[13] = 40

    helper(vm, "CALL(R12, R13)")

    assert vm.registers[13] == 101


def test_CALL_updates_frame_pointer(vm):
    vm.registers[12] = 600
    vm.registers[13] = 40

    helper(vm, "CALL(R12, R13)")

    assert vm.registers[14] == 600


def test_CALL_updates_first_register(vm):
    vm.registers[14] = 550
    vm.registers[12] = 600
    vm.registers[13] = 40

    helper(vm, "CALL(R12, R13)")

    assert vm.registers[12] == 550


def test_RETURN_changes_pc(vm):
    vm.pc = 100
    vm.registers[13] = 40

    helper(vm, "RETURN(R12, R13)")

    assert vm.pc == 40


def test_RETURN_updates_second_register(vm):
    vm.pc = 100
    vm.registers[13] = 40

    helper(vm, "RETURN(R12, R13)")

    assert vm.registers[13] == 101


def test_RETURN_updates_frame_pointer(vm):
    vm.registers[12] = 600
    vm.registers[13] = 40

    helper(vm, "RETURN(R12, R13)")

    assert vm.registers[14] == 600


def test_RETURN_updates_first_register(vm):
    vm.registers[14] = 550
    vm.registers[12] = 600
    vm.registers[13] = 40

    helper(vm, "RETURN(R12, R13)")

    assert vm.registers[12] == 550



"""
The parser for the HERA language.

Abstract grammar:
  start := (op | include)*

  op      := SYMBOL LPAREN arglist? RPAREN
  include := INCLUDE (STRING | BRACKETED)

  arglist := (value COMMA)* value


Author:  Ian Fisher (iafisher@protonmail.com)
Version: February 2019
"""
import os.path
from typing import List, Optional, Set, Tuple, Union  # noqa: F401

from .data import HERAError, Messages, Settings, Token
from .lexer import Lexer
from .op import AbstractOperation, name_to_class
from .stdlib import (
    TIGER_STDLIB_REG,
    TIGER_STDLIB_REG_DATA,
    TIGER_STDLIB_STACK,
    TIGER_STDLIB_STACK_DATA,
)
from .utils import read_file, register_to_index


def parse(
    text: str, *, path=None, settings=Settings()
) -> Tuple[List[AbstractOperation], Messages]:
    """
    Parse a HERA program.

    `path` is the path of the file being parsed, as it will appear in error and
    warning messages. It defaults to "<string>".
    """
    lexer = Lexer(text, path=path)
    parser = Parser(lexer, settings)
    program = parser.parse()
    return (program, parser.messages)


class Parser:
    def __init__(self, lexer: Lexer, settings: Settings) -> None:
        self.lexer = lexer
        self.visited = set()  # type: Set[str]
        self.settings = settings
        self.messages = Messages()

    def parse(self) -> List[AbstractOperation]:
        if self.lexer.path:
            self.visited.add(get_canonical_path(self.lexer.path))

        try:
            ops = self.match_program()
        except HERAError as e:
            self.messages.err(*e.args)
            return []

        # Make sure to capture any warnings from the lexer.
        self.messages.extend(self.lexer.messages)
        return ops

    def match_program(self) -> List[AbstractOperation]:
        expecting_brace = False
        ops = []
        while self.lexer.tkn.type != Token.EOF:
            msg = "expected HERA operation or #include"
            if not self.expect({Token.INCLUDE, Token.SYMBOL, Token.RBRACE}, msg):
                self.skip_until({Token.INCLUDE, Token.SYMBOL})
                continue

            if self.lexer.tkn.type == Token.INCLUDE:
                ops.extend(self.match_include())
            elif self.lexer.tkn.type == Token.SYMBOL:
                name_tkn = self.lexer.tkn
                self.lexer.next_token()
                # Many legacy HERA program are enclosed with void HERA_main() { ... },
                # which is handled here.
                if self.lexer.tkn.type == Token.SYMBOL and name_tkn.value == "void":
                    expecting_brace = True
                    self.handle_cpp_boilerplate()
                elif self.lexer.tkn.type == Token.LPAREN:
                    op = self.match_op(name_tkn)
                    if op:
                        ops.append(op)
                    # Operations may optionally be separated by semicolons.
                    if self.lexer.tkn.type == Token.SEMICOLON:
                        self.lexer.next_token()
                else:
                    self.err("expected left parenthesis")
            else:
                if expecting_brace:
                    expecting_brace = False
                else:
                    self.err("unexpected right brace")
                self.lexer.next_token()

        return ops

    def match_op(self, name_tkn: Token) -> Optional[AbstractOperation]:
        """
        Match an operation, assuming that self.lexer.tkn is on the left parenthesis.
        """
        self.lexer.next_token()
        args = self.match_optional_arglist()
        self.lexer.next_token()
        if args is None:
            return None

        try:
            cls = name_to_class[name_tkn.value]
        except KeyError:
            self.err("unknown instruction `{}`".format(name_tkn.value), name_tkn)
            return None
        else:
            return cls(*args, loc=name_tkn)

    VALUE_TOKENS = {
        Token.INT,
        Token.REGISTER,
        Token.SYMBOL,
        Token.STRING,
        Token.CHAR,
        Token.MINUS,
    }

    def match_optional_arglist(self) -> Optional[List[Token]]:
        """
        Match zero or more comma-separated values. Exits with the right parenthesis as
        the current token. Make sure to distinguish between a None return value (the
        arglist could not be parsed) and a [] return value (an empty arglist was parsed
        successfully).
        """
        if self.lexer.tkn.type == Token.RPAREN:
            return []

        args = []
        hit_error = False
        while True:
            if self.expect(self.VALUE_TOKENS, "expected value"):
                val = self.match_value()
            else:
                val = None

            if val is None:
                hit_error = True
                self.skip_until({Token.COMMA, Token.RPAREN})
                if self.lexer.tkn.type == Token.COMMA:
                    self.lexer.next_token()
                    continue
                else:
                    break
            else:
                args.append(val)

            self.lexer.next_token()
            if self.lexer.tkn.type == Token.RPAREN:
                break
            elif self.lexer.tkn.type != Token.COMMA:
                hit_error = True
                self.err("expected comma or right parenthesis")
                self.skip_until({Token.COMMA, Token.RPAREN})
                if (
                    self.lexer.tkn.type == Token.EOF
                    or self.lexer.tkn.type == Token.RPAREN
                ):
                    break
            else:
                self.lexer.next_token()

        return args if not hit_error else None

    def match_value(self) -> Optional[Token]:
        if self.lexer.tkn.type == Token.INT:
            return self.match_int()
        elif self.lexer.tkn.type == Token.CHAR:
            return Token(
                Token.INT, ord(self.lexer.tkn.value), location=self.lexer.tkn.location
            )
        elif self.lexer.tkn.type == Token.REGISTER:
            try:
                i = register_to_index(self.lexer.tkn.value)
            except HERAError:
                self.err("{} is not a valid register".format(self.lexer.tkn.value))
                i = 0
            self.lexer.tkn.value = i
            return self.lexer.tkn
        elif self.lexer.tkn.type == Token.MINUS:
            self.lexer.next_token()
            if not self.expect(Token.INT, "expected integer"):
                return None
            else:
                ret = self.match_int()
                ret.value *= -1
                return ret
        else:
            return self.lexer.tkn

    def match_int(self) -> Token:
        # Detect zero-prefixed octal numbers.
        prefix = self.lexer.tkn.value[:2]
        if len(prefix) == 2 and prefix[0] == "0" and prefix[1].isdigit():
            base = 8
            if self.settings.warn_octal_on:
                self.warn('consider using "0o" prefix for octal numbers')
        else:
            base = 0

        try:
            arg_as_int = int(self.lexer.tkn.value, base=base)
        except ValueError:
            self.err("invalid integer literal")
            # 1 is a neutral value that is valid anywhere an integer is.
            arg_as_int = 1
        self.lexer.tkn.value = arg_as_int
        return self.lexer.tkn

    def match_include(self) -> List[AbstractOperation]:
        root_path = self.lexer.path
        tkn = self.lexer.next_token()
        msg = "expected quote or angle-bracket delimited string"
        if not self.expect({Token.STRING, Token.BRACKETED}, msg):
            self.lexer.next_token()
            return []

        self.lexer.next_token()
        if tkn.type == Token.STRING:
            include_path = os.path.join(os.path.dirname(root_path), tkn.value)

            if get_canonical_path(include_path) in self.visited:
                self.err("recursive include", tkn)
                return []

            try:
                included_text = read_file(include_path)
            except HERAError as e:
                self.err(str(e), tkn)
                return []
            else:
                old_lexer = self.lexer
                self.lexer = Lexer(included_text, path=include_path)
                ops = self.parse()
                self.lexer = old_lexer
                return ops
        else:
            return self.expand_angle_include(tkn)

    def handle_cpp_boilerplate(self) -> None:
        self.lexer.next_token()
        if self.expect(Token.LPAREN, "expected left parenthesis"):
            self.lexer.next_token()

        if self.expect(Token.RPAREN, "expected right parenthesis"):
            self.lexer.next_token()

        self.expect(Token.LBRACE, "expected left curly brace")
        self.lexer.next_token()

    def expand_angle_include(self, include_path: Token) -> List[AbstractOperation]:
        # There is no check for recursive includes in this function, under the
        # assumption that system libraries do not have recursive includes.
        if include_path.value == "HERA.h":
            self.warn("#include <HERA.h> is not necessary for hera-py", include_path)
            return []
        elif include_path.value == "Tiger-stdlib-stack-data.hera":
            included_text = TIGER_STDLIB_STACK_DATA
        elif include_path.value == "Tiger-stdlib-stack.hera":
            included_text = TIGER_STDLIB_STACK
        elif include_path.value == "Tiger-stdlib-reg-data.hera":
            included_text = TIGER_STDLIB_REG_DATA
        elif include_path.value == "Tiger-stdlib-reg.hera":
            included_text = TIGER_STDLIB_REG
        else:
            root_path = os.environ.get("HERA_C_DIR", "/home/courses/lib/HERA-lib")
            try:
                included_text = read_file(os.path.join(root_path, include_path.value))
            except HERAError as e:
                self.err(str(e), include_path)
                return []

        old_lexer = self.lexer
        self.lexer = Lexer(included_text, path=include_path.value)
        ops = self.parse()
        self.lexer = old_lexer
        return ops

    def expect(self, types: Union[str, Set[str]], msg="unexpected token") -> bool:
        if isinstance(types, str):
            types = {types}

        if self.lexer.tkn.type not in types:
            if self.lexer.tkn.type == Token.EOF:
                self.err("premature end of input")
            elif self.lexer.tkn.type == Token.ERROR:
                self.err(self.lexer.tkn.value)
            else:
                self.err(msg)

            return False
        else:
            return True

    def skip_until(self, types: Set[str]) -> None:
        types.add(Token.EOF)
        while self.lexer.tkn.type not in types:
            self.lexer.next_token()

    def err(self, msg: str, tkn: Optional[Token] = None) -> None:
        if tkn is None:
            tkn = self.lexer.tkn
        self.messages.err(msg, tkn.location)

    def warn(self, msg: str, tkn: Optional[Token] = None) -> None:
        if tkn is None:
            tkn = self.lexer.tkn
        self.messages.warn(msg, tkn.location)


def get_canonical_path(fpath: str) -> str:
    if fpath == "-" or fpath == "<string>":
        return fpath
    else:
        return os.path.realpath(fpath)



"""
Type-checking and pseudo-op conversion.

Note that the code for type-checking individual operations lives in hera/op.py. This
module contains the code for program-wide type-checking.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: Feburary 2019
"""
from contextlib import suppress
from typing import Dict, List, Optional, Set, Tuple  # noqa: F401

from .data import (
    Constant,
    DataLabel,
    DebugInfo,
    Label,
    Messages,
    Program,
    Settings,
    Token,
)
from .op import (
    DataOperation,
    DebuggingOperation,
    AbstractOperation,
    RegisterBranch,
    RelativeBranch,
    LABEL,
    RTI,
    SWI,
)


def check(
    oplist: List[AbstractOperation], settings: Settings
) -> Tuple[Program, Messages]:
    symbol_table, messages = typecheck(oplist, settings=settings)
    if messages.errors:
        return (Program([], [], {}, None), messages)

    debug_info = None  # type: Optional[DebugInfo]
    if settings.mode == "debug":
        labels = labels_to_line_numbers(oplist)
        debug_info = DebugInfo(labels)
    else:
        debug_info = None

    oplist, preprocess_messages = convert_ops(oplist, symbol_table)
    messages.extend(preprocess_messages)

    data = []
    code = []
    for op in oplist:
        if isinstance(op, DataOperation):
            data.append(op)
        else:
            code.append(op)

    return (Program(data, code, symbol_table, debug_info), messages)


def typecheck(
    program: List[AbstractOperation], settings=Settings()
) -> Tuple[Dict[str, int], Messages]:
    """
    Type-check the program and emit error messages as appropriate. Return the
    program's symbol table.
    """
    messages = check_symbol_redeclaration(program)

    seen_code = False
    symbol_table, label_messages = get_labels(program, settings)
    messages.extend(label_messages)

    for op in program:
        op_messages = op.typecheck(symbol_table)
        messages.extend(op_messages)

        if isinstance(op, DataOperation):
            if seen_code:
                messages.err("data statement after code", loc=op.loc)
        else:
            seen_code = True

        if not settings.allow_interrupts and isinstance(op, (RTI, SWI)):
            messages.err("hera-py does not support {}".format(op.name), loc=op.loc)

        if settings.no_debug_ops and isinstance(op, DebuggingOperation):
            messages.err(
                "debugging instructions disallowed with --no-debug-ops flag", loc=op.loc
            )

        # Add constants to the symbol table as they are encountered, so that a given
        # constant is not in scope until after its declaration.
        if looks_like_a_CONSTANT(op):
            if out_of_range(op.args[1]):
                symbol_table[op.args[0]] = Constant(0)
            else:
                symbol_table[op.args[0]] = Constant(op.args[1])

    return (symbol_table, messages)


def check_symbol_redeclaration(program: List[AbstractOperation]) -> Messages:
    """
    Check if any symbols are redeclared in the program and return the error
    messages.
    """
    messages = Messages()
    symbols = set()  # type: Set[str]
    for op in program:
        if op.name in ("CONSTANT", "LABEL", "DLABEL") and len(op.args) >= 1:
            symbol = op.args[0]
            if symbol in symbols:
                messages.err(
                    "symbol `{}` has already been defined".format(symbol), loc=op.loc
                )
            else:
                symbols.add(symbol)
    return messages


def get_labels(
    program: List[AbstractOperation], settings: Settings
) -> Tuple[Dict[str, int], Messages]:
    """
    Return a dictionary mapping the labels and data labels (but not the constants) of
    the program to their concrete values.
    """
    messages = Messages()
    symbol_table = {}  # type: Dict[str, int]
    # We need to maintain a separate dictionary of constants because DSKIP can take a
    # constant as its argument, which has to be resolved to set the data counter
    # correctly.
    constants = {}
    pc = 0
    dc = settings.data_start
    for op in program:
        odc = dc
        if op.name == "LABEL":
            if len(op.args) == 1:
                symbol_table[op.args[0]] = Label(pc)
        elif op.name == "DLABEL":
            if len(op.args) == 1:
                if out_of_range(dc):
                    # If data counter has overflowed, put in a dummy value to avoid
                    # further overflow error messages.
                    symbol_table[op.args[0]] = DataLabel(0)
                else:
                    symbol_table[op.args[0]] = DataLabel(dc)
        elif op.name == "CONSTANT":
            if len(op.args) == 2:
                with suppress(ValueError):
                    constants[op.args[0]] = Constant(op.args[1])
        elif op.name == "INTEGER":
            dc += 1
        elif op.name == "LP_STRING" or op.name == "TIGER_STRING":
            if len(op.args) == 1 and isinstance(op.args[0], str):
                dc += len(op.args[0]) + 1
        elif op.name == "DSKIP":
            if len(op.args) == 1:
                if isinstance(op.args[0], int):
                    dc += op.args[0]
                elif op.args[0] in constants:
                    dc += constants[op.args[0]]
        else:
            pc += operation_length(op)

        # TODO: Can I move this elsewhere?
        if out_of_range(dc) and not out_of_range(odc):
            messages.err("past the end of available memory", loc=op.loc)

    return (symbol_table, messages)


def operation_length(op: AbstractOperation) -> int:
    if isinstance(op, RegisterBranch):
        if len(op.tokens) == 1 and op.tokens[0].type == Token.SYMBOL:
            return 3
        else:
            return 1
    elif op.name == "SET":
        return 2
    elif op.name == "CMP":
        return 2
    elif op.name == "SETRF":
        return 4
    elif op.name == "FLAGS":
        return 2
    elif op.name == "CALL":
        if len(op.tokens) == 2 and op.tokens[1].type != Token.REGISTER:
            return 3
        else:
            return 1
    elif op.name == "NEG":
        return 2
    elif op.name == "NOT":
        return 3
    else:
        return 1


def looks_like_a_CONSTANT(op: AbstractOperation) -> bool:
    return (
        op.name == "CONSTANT"
        and len(op.args) == 2
        and isinstance(op.args[0], str)
        and isinstance(op.args[1], int)
    )


def out_of_range(n: int) -> bool:
    return n < -32768 or n >= 65536


def convert_ops(
    oplist: List[AbstractOperation], symbol_table: Dict[str, int]
) -> Tuple[List[AbstractOperation], Messages]:
    """
    Convert the operations from pseudo-ops to real ops, and substitute values for
    labels and constants.

    The program must be type-checked before being passed to this function.
    """
    messages = Messages()
    retlist = []
    pc = 0
    for op in oplist:
        if isinstance(op, RelativeBranch) and op.tokens[0].type == Token.SYMBOL:
            target = symbol_table[op.args[0]]
            jump = target - pc
            if jump < -128 or jump >= 128:
                messages.err("label is too far for a relative branch", loc=op.tokens[0])
            else:
                op.tokens[0] = Token.Int(jump, location=op.tokens[0].location)
                op.args[0] = jump
        else:
            op = substitute_label(op, symbol_table)

        new_ops = op.convert()
        for new_op in new_ops:
            new_op.loc = op.loc
            new_op.original = op
            retlist.append(new_op)

        if not isinstance(op, DataOperation):
            pc += len(new_ops)
    return (retlist, messages)


def substitute_label(
    op: AbstractOperation, symbol_table: Dict[str, int]
) -> AbstractOperation:
    """Substitute any label in the instruction with its concrete value."""
    for i, tkn in enumerate(op.tokens):
        if tkn.type == Token.SYMBOL:
            op.tokens[i] = Token.Int(symbol_table[tkn.value], location=tkn.location)
            op.args[i] = symbol_table[tkn.value]
    return op


def labels_to_line_numbers(oplist: List[AbstractOperation]) -> Dict[str, str]:
    """
    Return a dictionary that maps from label names to their locations in the program,
    as a string of the form "<filepath>:<lineno>".
    """
    labels = {}
    for op in oplist:
        if isinstance(op, LABEL):
            labels[op.args[0]] = "{0.path}:{0.line}".format(op.loc)
    return labels



from io import StringIO
from unittest.mock import patch

from hera.main import main


def execute_program_helper(program, *, flags=[]):
    with patch("sys.stdin", StringIO(program)):
        return main(flags + ["--no-color", "-"])


def preprocess_program_helper(program):
    with patch("sys.stdin", StringIO(program)):
        main(["preprocess", "--no-color", "-"])



import pytest
from .utils import helper

from hera.data import DEFAULT_DATA_START
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_INTEGER_fills_in_memory(vm):
    helper(vm, "INTEGER(42)")

    assert vm.memory[DEFAULT_DATA_START] == 42


def test_INTEGER_increments_data_counter(vm):
    helper(vm, "INTEGER(42)")

    assert vm.dc == DEFAULT_DATA_START + 1


def test_INTEGER_does_not_increment_pc(vm):
    helper(vm, "INTEGER(42)")

    assert vm.pc == 0


def test_DSKIP_increments_data_counter(vm):
    helper(vm, "DSKIP(10)")

    assert vm.dc == DEFAULT_DATA_START + 10


def test_DSKIP_does_not_increment_pc(vm):
    helper(vm, "DSKIP(10)")

    assert vm.pc == 0


def test_LP_STRING_fills_in_memory(vm):
    helper(vm, 'LP_STRING("hello")')

    assert vm.memory[DEFAULT_DATA_START] == 5
    assert vm.memory[DEFAULT_DATA_START + 1] == ord("h")
    assert vm.memory[DEFAULT_DATA_START + 2] == ord("e")
    assert vm.memory[DEFAULT_DATA_START + 3] == ord("l")
    assert vm.memory[DEFAULT_DATA_START + 4] == ord("l")
    assert vm.memory[DEFAULT_DATA_START + 5] == ord("o")


def test_LP_STRING_increments_data_counter(vm):
    helper(vm, 'LP_STRING("hello")')

    assert vm.dc == DEFAULT_DATA_START + 6


def test_LP_STRING_does_not_increment_pc(vm):
    helper(vm, 'LP_STRING("hello")')

    assert vm.pc == 0



"""
The HERA assembler.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: March 2019
"""
import textwrap
from typing import List, Tuple

from .data import Program, Settings


def assemble(program: Program) -> Tuple[List[bytes], List[bytes]]:
    code = []
    data = []

    for op in program.code:
        code.append(op.assemble())

    for data_op in program.data:
        data.append(data_op.assemble())

    return (code, data)


def assemble_and_print(program: Program, settings: Settings) -> None:
    raw_code, raw_data = assemble(program)

    code = "\n".join(bytes_to_hex(b) for b in raw_code)

    raw_data_concat = b"".join(raw_data)
    datalist = []
    for i in range(0, len(raw_data_concat), 2):
        hi = raw_data_concat[i]
        lo = raw_data_concat[i + 1]
        datalist.append("{:x}".format((hi << 8) + lo))
    data = "\n".join(datalist)
    # I don't know what the significance of this cell is, but Hassem includes it.
    cell = (len(raw_data_concat) // 2) + settings.data_start
    data = "{:x}\n".format(cell) + data
    # Make sure to put zeroes up to the start of the data segment.
    data = "{}*0\n".format(settings.data_start - 1) + data

    if settings.stdout:
        if settings.data:
            print(data)
        elif settings.code:
            print(code)
        else:
            print("[DATA]")
            print(textwrap.indent(data, "  "))
            print("[CODE]")
            print(textwrap.indent(code, "  "))
    else:
        if settings.path == "-":
            path = "stdin"
        else:
            path = settings.path

        with open(path + ".lcode", "w", encoding="ascii") as f:
            f.write(code)
            f.write("\n")

        with open(path + ".ldata", "w", encoding="ascii") as f:
            f.write(data)


def bytes_to_hex(b: bytes) -> str:
    try:
        return b.hex()
    except AttributeError:
        # bytes.hex is not implemented in Python 3.4.
        return "".join("{:0>2x}".format(c) for c in b)



"""
The lexer for HERA and the debugging mini-language.

Consumed by hera/parser.py and hera/debugger/miniparser.py.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: February 2019
"""
import string
from typing import Optional, Tuple

from hera.data import Location, Messages, Token
from hera.utils import NAMED_REGISTERS


class Lexer:
    """A lexer for HERA (and for the debugging mini-language)."""

    def __init__(self, text: str, *, path: Optional[str] = None) -> None:
        self.text = text
        self.file_lines = text.splitlines()
        if self.text.endswith("\n"):
            self.file_lines.append("")
        self.position = 0
        self.line = 1
        self.column = 1
        self.path = path or "<string>"
        self.messages = Messages()
        # Set the current token.
        self.next_token()

    def get_location(self) -> Location:
        return Location(self.line, self.column, self.path, self.file_lines)

    def next_token(self) -> Token:
        self.skip()

        if self.position >= len(self.text):
            self.set_token(Token.EOF, length=0)
        else:
            ch = self.text[self.position]
            if ch.isalpha() or ch == "_":
                length = self.read_symbol()
                if is_register(self.text[self.position : self.position + length]):
                    self.set_token(Token.REGISTER, length=length)
                else:
                    self.set_token(Token.SYMBOL, length=length)
            elif ch.isdigit():
                length = self.read_int()
                self.set_token(Token.INT, length=length)
            elif ch == '"':
                self.consume_str()
            elif ch == "'":
                self.consume_char()
            elif self.text[self.position :].startswith("#include"):
                self.set_token(Token.INCLUDE, length=len("#include"))
            elif ch == "<":
                self.consume_bracketed()
            elif ch == ":":
                self.position += 1
                length = self.read_symbol()
                self.set_token(Token.FMT, length=length)
            elif ch == "-":
                self.set_token(Token.MINUS)
            elif ch == "+":
                self.set_token(Token.PLUS)
            elif ch == "/":
                self.set_token(Token.SLASH)
            elif ch == "*":
                self.set_token(Token.ASTERISK)
            elif ch == "@":
                self.set_token(Token.AT)
            elif ch == "(":
                self.set_token(Token.LPAREN)
            elif ch == ")":
                self.set_token(Token.RPAREN)
            elif ch == "{":
                self.set_token(Token.LBRACE)
            elif ch == "}":
                self.set_token(Token.RBRACE)
            elif ch == ",":
                self.set_token(Token.COMMA)
            elif ch == ";":
                self.set_token(Token.SEMICOLON)
            else:
                self.set_token(Token.UNKNOWN)

        return self.tkn

    def read_int(self) -> int:
        length = 1
        digits = {str(i) for i in range(10)}
        peek = self.peek_char()
        if self.text[self.position] == "0" and peek and peek in "boxBOX":
            length = 2
            if self.peek_char() in "xX":
                digits |= set(string.ascii_letters)

        while self.peek_char(length) in digits:
            length += 1

        return length

    def read_symbol(self) -> int:
        length = 1
        while True:
            ch = self.peek_char(length)
            if not (ch.isalpha() or ch.isdigit() or ch == "_"):
                break
            length += 1
        return length

    HEX_DIGITS = "0123456789abcdefABCDEF"

    def read_escape_char(self) -> Tuple[str, int]:
        """
        Read an escape sequence (assuming `self.text[self.position] == "\\"`) and return
        a pair (value, length), where `value` is what the escape sequence resolves to
        and `length` is the number of characters read.
        """
        peek = self.peek_char()
        loc = self.get_location()
        loc = loc._replace(column=loc.column + 1)
        if peek == "":
            return ("", 0)
        elif peek == "x":
            # Hex escapes
            peek2 = self.peek_char(2)
            peek3 = self.peek_char(3)
            if peek2 in self.HEX_DIGITS and peek3 in self.HEX_DIGITS:
                ordv = int(peek2 + peek3, base=16)
                return (chr(ordv), 3)
            else:
                self.warn("invalid hex escape", loc)
                return ("x", 1)
        elif peek.isdigit():
            # Octal escapes
            length = 1
            while length <= 3:
                if not self.peek_char(length).isdigit():
                    break
                length += 1

            val = self.text[self.position + 1 : self.position + length]

            try:
                ordv = int(val, base=8)
            except ValueError:
                self.warn("invalid octal escape", loc)
                return (self.peek_char(), 1)
            else:
                return (chr(ordv), length - 1)
        else:
            escape = escape_char(peek)
            if len(escape) == 2:
                self.warn("unrecognized backslash escape", loc)
            return (escape, 1)

    def consume_bracketed(self) -> None:
        self.next_char()
        loc = self.get_location()
        start = self.position
        while self.position < len(self.text) and self.text[self.position] != ">":
            self.next_char()

        if self.position == len(self.text):
            self.tkn = Token(Token.ERROR, "unclosed bracketed expression", loc)
            return

        self.tkn = Token(Token.BRACKETED, self.text[start : self.position], loc)
        self.next_char()

    def consume_str(self) -> None:
        loc = self.get_location()
        self.next_char()
        s = self.consume_delimited('"')

        if self.position == len(self.text):
            self.tkn = Token(Token.ERROR, "unclosed string literal", loc)
            return

        self.next_char()
        self.tkn = Token(Token.STRING, s, loc)

    def consume_char(self) -> None:
        loc = self.get_location()
        self.next_char()
        s = self.consume_delimited("'")

        if self.position == len(self.text):
            self.tkn = Token(Token.ERROR, "unclosed character literal", loc)
            return

        self.next_char()

        if len(s) == 1:
            self.tkn = Token(Token.CHAR, s, loc)
        elif len(s) == 2 and s[0] == "\\":
            self.tkn = Token(Token.CHAR, s[1], loc)
        else:
            self.tkn = Token(Token.ERROR, "over-long character literal", loc)

    def consume_delimited(self, delimiter) -> str:
        sbuilder = []
        while self.position < len(self.text) and self.text[self.position] != delimiter:
            if self.text[self.position] == "\\":
                value, length = self.read_escape_char()
                self.next_char()
                if length == 0:
                    # Length of zero indicates EOF.
                    break

                sbuilder.append(value)
                for _ in range(length):
                    self.next_char()
            else:
                sbuilder.append(self.text[self.position])
                self.next_char()

        return "".join(sbuilder)

    def skip(self) -> None:
        """Skip past whitespace and comments."""
        while True:
            # Skip whitespace.
            while self.position < len(self.text) and self.text[self.position].isspace():
                self.next_char()

            # Skip comments.
            if self.position < len(self.text) and self.text[self.position] == "/":
                if self.peek_char() == "/":
                    while self.position < len(self.text):
                        if self.text[self.position] == "\n":
                            break
                        self.next_char()
                elif self.peek_char() == "*":
                    # Move past the "/*" at the start of the comment.
                    self.next_char()
                    self.next_char()
                    while self.position < len(self.text):
                        if self.text[self.position] == "*" and self.peek_char() == "/":
                            break
                        self.next_char()
                    # Skip the "*/" at the end of the comment.
                    self.next_char()
                    self.next_char()
                else:
                    break
            else:
                break

    def next_char(self) -> None:
        """
        Advance the position in the text by one. Do not call this method if the current
        position is past the end of the text.
        """
        if self.text[self.position] == "\n":
            self.line += 1
            self.column = 1
        else:
            self.column += 1
        self.position += 1

    def peek_char(self, n=1) -> str:
        """
        Return the n'th character in the text past the current position. If past the
        end, return the empty string.
        """
        return (
            self.text[self.position + n] if self.position + n < len(self.text) else ""
        )

    def set_token(self, typ: str, *, length=1) -> None:
        loc = self.get_location()
        value = self.text[self.position : self.position + length]
        for _ in range(length):
            self.next_char()
        self.tkn = Token(typ, value, loc)

    def warn(self, msg: str, loc) -> None:
        self.messages.warn(msg, loc)


def escape_char(c):
    """
    Return the special character that `c` encodes.

        >>> escape_char("n")
        "\n"
    """
    if c == "n":
        return "\n"
    elif c == "t":
        return "\t"
    elif c == "\\":
        return "\\"
    elif c == '"':
        return '"'
    else:
        return "\\" + c


def is_register(s):
    return (s[0] in "rR" and s[1:].isdigit()) or s.lower() in NAMED_REGISTERS



import pytest
from .utils import helper

from hera.utils import to_u16
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_SETLO_with_positive(vm):
    helper(vm, "SETLO(R5, 23)")

    assert vm.registers[5] == 23


def test_SETLO_with_negative(vm):
    helper(vm, "SETLO(R9, -12)")

    assert vm.registers[9] == to_u16(-12)


def test_SETLO_with_max_positive(vm):
    helper(vm, "SETLO(R2, 127)")

    assert vm.registers[2] == 127


def test_SETLO_with_255(vm):
    helper(vm, "SETLO(R2, 255)")

    assert vm.registers[2] == to_u16(-1)


def test_SETLO_with_max_negative(vm):
    helper(vm, "SETLO(R2, -128)")

    assert vm.registers[2] == to_u16(-128)


def test_SETLO_clears_high_bits(vm):
    vm.registers[6] = 4765

    helper(vm, "SETLO(R6, 68)")

    assert vm.registers[6] == 68


def test_SETLO_increments_pc(vm):
    helper(vm, "SETLO(R9, -12)")

    assert vm.pc == 1


def test_SETLO_ignores_flags(vm):
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.flag_sign = True
    vm.flag_zero = False

    helper(vm, "SETLO(R7, 0)")

    assert vm.flag_carry
    assert vm.flag_overflow
    assert vm.flag_sign
    assert not vm.flag_zero


def test_SETLO_does_not_set_zero_flag(vm):
    helper(vm, "SETLO(R7, 0)")

    assert not vm.flag_zero


def test_SETLO_does_not_set_sign_flag(vm):
    helper(vm, "SETLO(R7, -1)")

    assert not vm.flag_sign


def test_SETLO_does_not_change_R0(vm):
    helper(vm, "SETLO(R0, 20)")

    assert vm.registers[0] == 0


def test_SETHI_with_positive(vm):
    helper(vm, "SETHI(R5, 23)")

    assert vm.registers[5] == 5888


def test_SETHI_with_max_positive(vm):
    helper(vm, "SETHI(R2, 255)")

    assert vm.registers[2] == 65280


def test_SETHI_does_not_clear_low_bits(vm):
    vm.registers[6] = 4765

    helper(vm, "SETHI(R6, 68)")

    assert vm.registers[6] == 17565


def test_SETHI_increments_pc(vm):
    helper(vm, "SETHI(R9, 12)")

    assert vm.pc == 1


def test_SETHI_ignores_flags(vm):
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.flag_sign = True
    vm.flag_zero = False

    helper(vm, "SETHI(R7, 0)")

    assert vm.flag_carry
    assert vm.flag_overflow
    assert vm.flag_sign
    assert not vm.flag_zero


def test_SETHI_does_not_set_zero_flag(vm):
    helper(vm, "SETHI(R7, 0)")

    assert not vm.flag_zero


def test_SETHI_does_not_set_sign_flag(vm):
    helper(vm, "SETHI(R7, -1)")

    assert not vm.flag_sign


def test_SETHI_does_not_change_R0(vm):
    helper(vm, "SETHI(R0, 20)")

    assert vm.registers[0] == 0


def test_INC_with_small_positive(vm):
    helper(vm, "INC(R8, 6)")

    assert vm.registers[8] == 6


def test_INC_with_max(vm):
    helper(vm, "INC(R2, 32)")

    assert vm.registers[2] == 32


def test_INC_with_previous_value(vm):
    vm.registers[5] = 4000

    helper(vm, "INC(R5, 2)")

    assert vm.registers[5] == 4002


def test_INC_with_previous_negative_value(vm):
    vm.registers[9] = to_u16(-12)

    helper(vm, "INC(R9, 10)")

    assert vm.registers[9] == to_u16(-2)


def test_INC_increments_pc(vm):
    helper(vm, "INC(R1, 1)")

    assert vm.pc == 1


def test_INC_sets_zero_flag(vm):
    vm.registers[7] = to_u16(-1)

    helper(vm, "INC(R7, 1)")

    assert vm.registers[7] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_INC_sets_sign_flag(vm):
    vm.registers[1] = 32765

    helper(vm, "INC(R1, 5)")

    assert vm.registers[1] == to_u16(-32766)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_INC_sets_carry_flag(vm):
    vm.registers[8] = to_u16(-1)

    helper(vm, "INC(R8, 1)")

    assert vm.flag_carry
    assert not vm.flag_overflow


def test_INC_sets_overflow_flag(vm):
    vm.registers[8] = 32765

    helper(vm, "INC(R8, 5)")

    assert not vm.flag_carry
    assert vm.flag_overflow


def test_INC_ignores_incoming_carry(vm):
    vm.flag_carry = True

    helper(vm, "INC(R8, 5)")

    assert vm.registers[8] == 5
    assert not vm.flag_carry


def test_INC_does_not_affect_R0(vm):
    helper(vm, "INC(R0, 1)")

    assert vm.registers[0] == 0


def test_DEC_with_small_positive(vm):
    helper(vm, "DEC(R8, 6)")

    assert vm.registers[8] == to_u16(-6)


def test_DEC_with_max(vm):
    helper(vm, "DEC(R2, 32)")

    assert vm.registers[2] == to_u16(-32)


def test_DEC_with_previous_value(vm):
    vm.registers[5] = 4000

    helper(vm, "DEC(R5, 2)")

    assert vm.registers[5] == 3998


def test_DEC_with_previous_negative_value(vm):
    vm.registers[9] = to_u16(-12)

    helper(vm, "DEC(R9, 10)")

    assert vm.registers[9] == to_u16(-22)


def test_DEC_increments_pc(vm):
    helper(vm, "DEC(R1, 1)")

    assert vm.pc == 1


def test_DEC_sets_zero_flag(vm):
    vm.registers[7] = 1

    helper(vm, "DEC(R7, 1)")

    assert vm.registers[7] == 0
    assert vm.flag_zero
    assert not vm.flag_sign


def test_DEC_sets_sign_flag(vm):
    vm.registers[1] = 1

    helper(vm, "DEC(R1, 5)")

    assert vm.registers[1] == to_u16(-4)
    assert not vm.flag_zero
    assert vm.flag_sign


def test_DEC_sets_carry_flag(vm):
    helper(vm, "DEC(R8, 1)")

    assert vm.flag_carry
    assert not vm.flag_overflow


def test_DEC_sets_overflow_flag(vm):
    vm.registers[8] = to_u16(-32768)

    helper(vm, "DEC(R8, 5)")

    assert vm.registers[8] == 32763
    assert not vm.flag_carry
    assert vm.flag_overflow


def test_DEC_ignores_incoming_carry(vm):
    vm.flag_carry = True
    vm.registers[8] = 10

    helper(vm, "DEC(R8, 5)")

    assert vm.registers[8] == 5
    assert not vm.flag_carry


def test_DEC_does_not_affect_R0(vm):
    helper(vm, "DEC(R0, 1)")

    assert vm.registers[0] == 0



from hera.checker import check
from hera.data import Settings
from hera.parser import parse


def helper(vm, opstr):
    # A little abstraction for the test suite, in case the virtual machine's API
    # changes.
    oplist, messages = parse(opstr)
    assert not messages.errors, messages.errors[0]
    program, messages = check(oplist, Settings())
    assert not messages.errors, messages.errors[0]

    # Take first op, or first data statement if there are no ops.
    op = (program.code + program.data)[0]
    op.execute(vm)



from hera.data import DEFAULT_DATA_START
from hera.main import main


def test_aslu_program(capsys):
    vm = main(["test/assets/cs240/aslu.hera"])

    assert vm.registers[1] == 0xBFF2
    assert vm.registers[2] == 0xF000
    assert vm.registers[3] == 0x0010
    assert vm.registers[4] == 0x7000
    assert vm.registers[5] == 0x8010
    assert vm.registers[6] == 0x7800
    assert vm.registers[7] == 0x0070
    assert vm.registers[8] == 0xE009
    assert vm.registers[9] == 0xFFC8
    assert vm.registers[10] == 0x3C00

    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block

    assert "Warning" not in capsys.readouterr().err


def test_branches_program():
    vm = main(["test/assets/cs240/branches.hera"])

    assert vm.registers[1] == 1
    assert vm.registers[2] == 2
    assert vm.registers[3] == 3
    assert vm.registers[4] == 4
    assert vm.registers[5] == 5
    assert vm.registers[6] == 6

    for r in vm.registers[7:11]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block

    # This code does give a warning on account of C++ boilerplate.


def test_fibonacci_program():
    vm = main(["test/assets/cs240/fib.hera"])

    assert vm.registers[1] == 0
    assert vm.registers[2] == 0
    assert vm.registers[3] == 0
    assert vm.registers[4] == 0x000A
    assert vm.registers[5] == 0x0037
    assert vm.registers[6] == 0x0022
    assert vm.registers[7] == 0x0022
    assert vm.registers[8] == 0x000B
    assert vm.registers[9] == 0
    assert vm.registers[10] == 0

    assert vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block

    # This code does give a warning on account of C++ boilerplate.


def test_flag_program():
    vm = main(["test/assets/cs240/flag.hera"])

    assert vm.registers[1] == 0x0015
    assert vm.registers[2] == 0x0000
    assert vm.registers[3] == 0x001B
    assert vm.registers[4] == 0x0009
    assert vm.registers[5] == 0x0019
    assert vm.registers[6] == 0x0003
    assert vm.registers[7] == 0x0003
    assert vm.registers[8] == 0x0015
    assert vm.registers[9] == 0x0000
    assert vm.registers[10] == 0x0007

    assert vm.flag_sign
    assert not vm.flag_zero
    assert vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block

    # This code does give a warning on account of C++ boilerplate.


def test_stein_program():
    vm = main(["test/assets/cs240/stein.hera"])

    assert vm.registers[1] == 1
    assert vm.registers[2] == 1
    assert vm.registers[3] == 2
    assert vm.registers[4] == 1

    for r in vm.registers[5:11]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block

    # This code does give a warning on account of C++ boilerplate.


def test_factorial_program(capsys):
    vm = main(["test/assets/cs240/factorial.hera"])

    assert vm.registers[1] == 7
    assert vm.registers[2] == 5040

    for r in vm.registers[3:11]:
        assert r == 0

    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert vm.flag_carry_block

    assert "Warning" not in capsys.readouterr().err


def test_extended_stein_program():
    vm = main(["test/assets/cs240/extended_stein.hera"])

    assert vm.registers[1] == 0x001
    assert vm.registers[2] == 0x001
    assert vm.registers[3] == 0x001
    assert vm.registers[4] == 0x0011
    assert vm.registers[5] == 0x0027
    assert vm.registers[6] == 0x0017
    assert vm.registers[7] == 0xFFF6
    assert vm.registers[8] == 0xFFF0
    assert vm.registers[9] == 0x0007
    assert vm.registers[10] == 0x0001

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block

    # This code does give a warning on account of C++ boilerplate.


def test_call_and_return_program(capsys):
    vm = main(["test/assets/cs240/call_and_return.hera"])

    assert vm.registers[1] == 0x0009
    assert vm.registers[2] == 0x0013

    for r in vm.registers[3:11]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert vm.flag_carry_block

    assert "Warning" not in capsys.readouterr().err


def test_array_program():
    vm = main(["test/assets/cs240/array.hera"])

    assert vm.registers[5] == 5050

    for r in vm.registers[6:11]:
        assert r == 0

    assert not vm.flag_sign
    assert vm.flag_zero
    assert not vm.flag_overflow
    assert vm.flag_carry
    assert not vm.flag_carry_block

    for i in range(100):
        assert vm.memory[DEFAULT_DATA_START + i] == i + 1
    assert vm.memory[DEFAULT_DATA_START + 100] == 5050



from hera.data import Token
from hera.op import ADD, INC, LABEL, RTI, SET
from hera.parser import parse


def valid(text, *, warnings=False):
    program, messages = parse(text)
    assert not messages.errors, messages.errors[0]
    if not warnings:
        assert not messages.warnings, messages.warnings[0]
    return program


def invalid(text):
    program, messages = parse(text)
    return messages


def test_parse_single_op():
    program = valid("SET(R1, 42)")

    assert program == [SET(R(1), INT(42))]


def test_parse_has_locations():
    program = valid("SET(  R1,\n42)")

    op = program[0]
    assert isinstance(op, SET)
    assert op.loc.line == 1
    assert op.loc.column == 1

    assert op.tokens[0].location.line == 1
    assert op.tokens[0].location.column == 7

    assert op.tokens[1].location.line == 2
    assert op.tokens[1].location.column == 1


def test_parse_op_with_no_args():
    program = valid("RTI()")

    assert program == [RTI()]


def test_parse_multiple_ops():
    program = valid("INC(R4, 5)\nADD(R1, R2, R3)")

    assert program == [INC(R(4), INT(5)), ADD(R(1), R(2), R(3))]


def test_parse_label_starting_with_register_name():
    program = valid("LABEL(R1_INIT)")

    assert program == [LABEL(Token.Sym("R1_INIT"))]


def test_parse_octal_number():
    program = valid("SETLO(R1, 0o123)")

    assert program[0].args[1] == 0o123


def test_parse_hex_number():
    program = valid("SETHI(R1, 0xaBc)")

    assert program[0].args[1] == 0xABC


def test_parse_binary_number():
    program = valid("INC(R7, 0b10101)")

    assert program[0].args[1] == 0b10101


def test_parse_empty_program():
    program = valid("")

    assert len(program) == 0


def test_parse_single_line_comment():
    program = valid("// A comment")

    assert len(program) == 0


def test_parse_single_line_comment_with_ops():
    program = valid("SETLO(R1, 1) // A comment\nSETHI(R1, 1)")

    assert len(program) == 2
    assert program[0].name == "SETLO"
    assert program[1].name == "SETHI"


def test_parse_cpp_boilerplate():
    program = valid(
        """\
#include <HERA.h>

void HERA_main() {
    SET(R1, 42)
}
    """,
        warnings=True,
    )

    assert len(program) == 1
    assert program[0].name == "SET"


def test_parse_cpp_boilerplate_weird_whitespace_and_spelling():
    program = valid("#include <HERA.h>\nvoid   HeRA_mAin( \t)\n {\n\n}", warnings=True)

    assert program == []


def test_parse_empty_cpp_boilerplate():
    program = valid(
        """\
#include <HERA.h>

void HERA_main() {
}
    """,
        warnings=True,
    )

    assert len(program) == 0


def test_parse_ops_with_semicolons():
    program = valid("SET(R1, 1); SET(R2, 2)")

    assert len(program) == 2
    assert program[0].name == "SET"
    assert program[1].name == "SET"


def SYM(s):
    return Token(Token.SYMBOL, s)


def INT(x):
    return Token(Token.INT, x)


def R(i):
    return Token(Token.REGISTER, i)



"""
Some important data structures for the HERA interpreter.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: February 2019
"""
from collections import namedtuple


DEFAULT_DATA_START = 0xC001

VOLUME_QUIET = "quiet"
VOLUME_NORMAL = "normal"
VOLUME_VERBOSE = "verbose"


class Settings:
    """Global settings of the interpreter."""

    def __init__(self, *, color=True, mode="", volume=VOLUME_NORMAL):
        self.allow_interrupts = False
        self.code = False
        self.color = color
        self.data = False
        self.data_start = DEFAULT_DATA_START
        self.mode = mode
        self.no_debug_ops = False
        self.obfuscate = False
        self.path = None
        self.stdout = False
        self.throttle = False
        self.warn_octal_on = True
        self.warn_return_on = True
        self.volume = volume
        self.warning_count = 0


class Location(namedtuple("Location", ["line", "column", "path", "file_lines"])):
    """
    A class to represent a location in a file, for the use of warnings and error
    messages. The `file_lines` field is a pointer to a list of lines in the file, so
    that the actual line can be printed in the message.
    """

    def __repr__(self):
        if self.file_lines:
            return "Location(line={}, column={}, path={!r}, file_lines=[...])".format(
                self.line, self.column, self.path
            )
        else:
            return "Location(line={}, column={}, path={!r}, file_lines=[])".format(
                self.line, self.column, self.path
            )


Program = namedtuple("Program", ["data", "code", "symbol_table", "debug_info"])
DebugInfo = namedtuple("DebugInfo", ["labels"])


class Token:
    # Possible values for the `type` field.
    INT = "TOKEN_INT"
    REGISTER = "TOKEN_REGISTER"
    SYMBOL = "TOKEN_SYMBOL"
    STRING = "TOKEN_STRING"
    BRACKETED = "TOKEN_BRACKETED"
    CHAR = "TOKEN_CHAR"
    MINUS = "TOKEN_MINUS"
    AT = "TOKEN_AT"
    ASTERISK = "TOKEN_ASTERISK"
    PLUS = "TOKEN_PLUS"
    SLASH = "TOKEN_SLASH"
    LPAREN = "TOKEN_LPAREN"
    RPAREN = "TOKEN_RPAREN"
    LBRACE = "TOKEN_LBRACE"
    RBRACE = "TOKEN_RBRACE"
    COMMA = "TOKEN_COMMA"
    SEMICOLON = "TOKEN_SEMICOLON"
    FMT = "TOKEN_FMT"
    INCLUDE = "TOKEN_INCLUDE"
    EOF = "TOKEN_EOF"
    ERROR = "TOKEN_ERROR"
    UNKNOWN = "TOKEN_UNKNOWN"

    def __init__(self, type_, value, location=None):
        self.type = type_
        self.value = value
        self.location = location

    @classmethod
    def R(cls, i, location=None):
        return cls(cls.REGISTER, i, location)

    @classmethod
    def Int(cls, x, location=None):
        return cls(cls.INT, x, location)

    @classmethod
    def Sym(cls, s, location=None):
        return cls(cls.SYMBOL, s, location)

    def __eq__(self, other):
        if isinstance(other, Token):
            return self.type == other.type and self.value == other.value
        else:
            return self.value == other

    def __repr__(self):
        return "Token({0.type!r}, {0.value!r}, loc={0.location!r})".format(self)


class Messages:
    def __init__(self, msg=None, loc=None, *, warning=False):
        self.errors = []
        self.warnings = []
        if msg is not None:
            if warning:
                self.warnings.append((msg, loc))
            else:
                self.errors.append((msg, loc))

    def extend(self, messages):
        self.errors.extend(messages.errors)
        self.warnings.extend(messages.warnings)
        return self

    def err(self, msg, loc=None):
        self.errors.append((msg, loc))

    def warn(self, msg, loc=None):
        self.warnings.append((msg, loc))


class HERAError(Exception):
    pass


class Constant(int):
    pass


class Label(int):
    pass


class DataLabel(int):
    pass



from .utils import preprocess_program_helper


def test_preprocess_simple_program(capsys):
    preprocess_program_helper("SET(R1, 10)")

    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        captured.out
        == """\

  0000  SETLO(R1, 10)
  0001  SETHI(R1, 0)
"""
    )


def test_preprocess_data(capsys):
    preprocess_program_helper('LP_STRING("hello")')

    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        captured.out
        == """\

[DATA]
  LP_STRING("hello")
"""
    )


def test_preprocess_data_and_code(capsys):
    preprocess_program_helper('DLABEL(s) LP_STRING("hello") SET(R1, s)')

    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        captured.out
        == """\

[DATA]
  LP_STRING("hello")

[CODE]
  0000  SETLO(R1, 1)
  0001  SETHI(R1, 192)
"""
    )


def test_preprocess_character_literal(capsys):
    preprocess_program_helper("SET(R1, 'A')")

    captured = capsys.readouterr()
    assert captured.err == ""
    assert captured.out == "\n  0000  SETLO(R1, 65)\n  0001  SETHI(R1, 0)\n"



"""
The shell interface to the HERA debugger.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: March 2019
"""
import copy
import functools
import textwrap
from typing import List, Optional

from . import miniparser
from .debugger import Debugger
from .miniparser import (
    InfixNode,
    IntNode,
    MemoryNode,
    PrefixNode,
    RegisterNode,
    SymbolNode,
)
from hera.assembler import assemble_and_print
from hera.data import DataLabel, HERAError, Label, Location, Program, Settings
from hera.loader import load_program
from hera.op import Branch, DataOperation, disassemble, LABEL, OPCODE
from hera.parser import parse
from hera.utils import format_int, pad


def debug(program: Program, settings: Settings) -> None:
    """Start the debug loop."""
    debugger = Debugger(program, settings)
    Shell(debugger, settings).loop()


def mutates(handler):
    """
    Decorator for command handlers in the Shell class that mutate the state of the
    debugger.
    """

    @functools.wraps(handler)
    def inner(self, *args, **kwargs):
        self.debugger.save()
        name = handler.__name__[len("handle_") :]
        self.command_history.append(name)
        return handler(self, *args, **kwargs)

    return inner


class Shell:
    def __init__(self, debugger: Debugger, settings: Settings) -> None:
        self.debugger = debugger
        self.settings = settings
        self.command_history = []  # type: List[str]

    def loop(self):
        if self.debugger.empty():
            print("Cannot debug an empty program.")
            return

        self.print_current_op()

        while True:
            try:
                response = input(">>> ").strip()
            except (EOFError, KeyboardInterrupt):
                print()
                break

            if not response:
                continue
            else:
                should_continue = self.handle_command(response)
                if should_continue is False:
                    break

    def handle_command(self, response: str) -> bool:
        """
        Parse the command and execute it. Return False if the loop should exit, and
        True otherwise.
        """
        try:
            cmd, argstr = response.split(maxsplit=1)
        except ValueError:
            cmd = response
            argstr = ""

        try:
            fullcmd = self.expand_command(cmd)
        except HERAError:
            if "=" in response:
                self.handle_assign(response.split("=", maxsplit=1))
            else:
                print("{} is not a recognized command.".format(cmd))

            return True
        else:
            if fullcmd == "quit":
                return False

            handler = getattr(self, "handle_" + fullcmd)
            if fullcmd in self.TAKES_ARGSTR:
                handler(argstr)
            else:
                handler(argstr.split())

            return True

    # Commands which may be abbreviated with a prefix. Order determines precedence when
    # multiple commands share a prefix.
    CAN_BE_ABBREVIATED = (
        # Multiple lists to prevent reformatting.
        ["assign", "break", "continue", "clear", "execute", "goto"]
        + ["help", "info", "list", "next", "print", "quit", "step", "undo"]
    )

    # Commands that require the whole command to be spelled out.
    CANNOT_BE_ABBREVIATED = ["asm", "dis", "ll", "off", "on", "restart"]

    # Commands that do not take a whitespace-separated list of argument.
    TAKES_ARGSTR = ["asm", "execute", "print"]

    def expand_command(self, cmd: str) -> str:
        cmd = cmd.lower()
        for full in self.CAN_BE_ABBREVIATED:
            if full.startswith(cmd):
                return full

        for full in self.CANNOT_BE_ABBREVIATED:
            if full == cmd:
                return full

        raise HERAError

    def handle_asm(self, argstr: str) -> None:
        """
        asm <op>
          Assemble the HERA operation into machine code
        """
        if not argstr.strip():
            print("asm takes one argument.")
            return

        # Check if the program is all code, all data, or a mix.
        any_code = False
        any_data = False
        for op in parse(argstr)[0]:
            if isinstance(op, DataOperation):
                any_data = True
            else:
                any_code = True

        try:
            program = load_program(argstr, self.settings)
        except SystemExit:
            return

        code_flag = any_code and not any_data
        data_flag = any_data and not any_code
        settings = self.asm_settings(code=code_flag, data=data_flag)
        assemble_and_print(program, settings)

    @mutates
    def handle_assign(self, args: List[str]) -> None:
        """
        assign <x> <y>
          Assign the value of y to x. x may be a register, a memory location, or the
          program counter. y may be a register, a memory location, the program counter,
          a symbol, or an integer.

        <x> = <y>
          Alias for "assign <x> <y>", with the additional advantage that <x> and <y>
          may contain spaces.

          Examples:
            Assign to a register:          R1 = 42
            Assign to a memory location:   @(1000) = R4
            Assign a label to a register:  R1 = some_label
            Arithmetic:                    R7 = R5 * 10
        """
        if len(args) != 2:
            print("assign takes two arguments.")
            return

        try:
            ltree = miniparser.parse(args[0])
            rtree = miniparser.parse(args[1])
        except SyntaxError as e:
            msg = str(e)
            if msg:
                print("Parse error: " + msg + ".")
            else:
                print("Parse error.")
            return

        if len(ltree.seq) > 1:
            print("Parse error: cannot assign to sequence.")
            return

        if len(rtree.seq) > 1:
            print("Parse error: cannot assign sequence value.")
            return

        ltree = ltree.seq[0]
        rtree = rtree.seq[0]

        vm = self.debugger.vm
        try:
            rhs = self.evaluate_node(rtree)
            if isinstance(ltree, RegisterNode):
                vm.store_register(ltree.value, rhs)
            elif isinstance(ltree, MemoryNode):
                address = self.evaluate_node(ltree.address)
                vm.store_memory(address, rhs)
            elif isinstance(ltree, SymbolNode):
                if ltree.value == "pc":
                    vm.pc = rhs
                else:
                    print("Eval error: cannot assign to symbol.")
            elif isinstance(ltree, (InfixNode, PrefixNode)):
                print("Eval error: cannot assign to arithmetic expression.")
            else:
                raise RuntimeError(
                    "unknown node type {}".format(ltree.__class__.__name__)
                )
        except HERAError as e:
            print("Eval error: " + str(e) + ".")

    @mutates
    def handle_break(self, args: List[str]) -> None:
        """
        break
          Print all current breakpoints.

        break <n>
          Set a breakpoint at the given line number in the file that the debugger was
          opened on.

        break <path>:<n>
          Set a breakpoint at the given line number in the given file.

        break <label>
          Set a breakpoint at the given label.
          
        break .
          Set a breakpoint at the current instruction.
        """
        if len(args) > 1:
            print("break takes zero or one arguments.")
            return

        if len(args) == 0:
            breakpoints = self.debugger.get_breakpoints()
            if breakpoints:
                for brk in breakpoints.values():
                    print(brk)
            else:
                print("No breakpoints set.")
        else:
            try:
                b = self.debugger.resolve_location(args[0])
            except ValueError as e:
                print("Error:", e)
            else:
                self.debugger.set_breakpoint(b)
                loc = self.debugger.op(b).loc
                print("Breakpoint set in file {0.path}, line {0.line}.".format(loc))

    @mutates
    def handle_clear(self, args: List[str]) -> None:
        """
        clear <location>
          Clear a breakpoint at the given line. Location formats accepted are the same
          as the break command.

        clear *
          Clear all breakpoints.
        """
        if len(args) == 0:
            print("clear takes one or more arguments.")
            return

        if any(arg == "*" for arg in args):
            self.debugger.breakpoints.clear()
            print("Cleared all breakpoints.")
        else:
            for arg in args:
                try:
                    b = self.debugger.resolve_location(arg)
                except ValueError as e:
                    print("Error:", e)
                else:
                    if b in self.debugger.breakpoints:
                        loc = self.debugger.op(b).loc
                        del self.debugger.breakpoints[b]
                        msg = "Cleared breakpoint in file {0.path}, line {0.line}."
                        print(msg.format(loc))
                    else:
                        print("No breakpoint at that location.")

    @mutates
    def handle_continue(self, args: List[str]) -> None:
        """
        continue
          Execute the program until a breakpoint is encountered or the program
          terminates.
        """
        if len(args) != 0:
            print("continue takes no arguments.")
            return

        self.debugger.next(step=True)
        while not self.debugger.finished() and not self.debugger.at_breakpoint():
            self.debugger.next(step=True)

        self.print_current_op()

    def handle_dis(self, args: List[str]) -> None:
        """
        dis <n>
          Interpret the 16-bit integer as a HERA machine instruction, and disassemble
          it into its assembly-language mnemonic.

        dis <n1> <n2>...
          Disassemble multiple integers.

        dis
          If the current instruction is an OPCODE, disassemble its contents.
        """
        if len(args) > 0:
            intargs = []
            for a in args:
                try:
                    v = int(a, base=0)
                except ValueError:
                    print("Could not parse argument `{}` to dis.".format(a))
                    return
                else:
                    intargs.append(v)

            for v in intargs:
                try:
                    print(disassemble(v))
                except HERAError as e:
                    print("Error:", e)
        else:
            if not self.debugger.finished() and isinstance(self.debugger.op(), OPCODE):
                op = self.debugger.op()
                print(disassemble(op.args[0]))
            else:
                print("Current operation is not an OPCODE.")

    @mutates
    def handle_execute(self, argstr: str) -> None:
        """
        execute <op>
          Execute a HERA operation. The operation must not be a data statement or a
          branch. The operation may affect registers and memory. Some operations can
          be more concisely expressed with the debugging mini-language. Type
          "help assign" for details.

          Examples:
            execute ASR(R5, R4)
            execute SET(R1, 20)  SET(R2, 22)  ADD(R3, R2, R1)
        """
        if not argstr.strip():
            print("execute takes one argument.")
            return

        # Make sure there are no disallowed ops.
        for op in parse(argstr)[0]:
            if isinstance(op, Branch):
                print("execute cannot take branching operations.")
                return
            elif isinstance(op, DataOperation):
                print("execute cannot take data statements.")
                return
            elif isinstance(op, LABEL):
                print("execute cannot take labels.")
                return

        try:
            program = load_program(argstr, self.settings)
        except SystemExit:
            return

        vm = self.debugger.vm
        opc = vm.pc
        for op in program.code:
            op.execute(vm)
        vm.pc = opc

    @mutates
    def handle_goto(self, args: List[str]) -> None:
        """
        goto <loc>
          Jump to the given location (either a line number or a label) without
          executing any of the intermediate instructions.
        """
        if len(args) != 1:
            print("goto takes one argument.")
            return

        try:
            new_pc = self.debugger.resolve_location(args[0])
        except ValueError as e:
            print("Error:", str(e))
            return
        else:
            self.debugger.vm.pc = new_pc

        self.print_current_op()

    def handle_help(self, args: List[str]) -> None:
        """
        help
          Print a summary of all debugging commands.

        help <cmd>...
          Print a detailed help message for each command list.
        """
        if not args:
            print(HELP)
        else:
            for i, arg in enumerate(args):
                try:
                    fullarg = self.expand_command(arg)
                except HERAError:
                    print("{} is not a recognized command.".format(arg))
                else:
                    if fullarg == "quit":
                        print("quit\n  Exit the debugger.")
                    else:
                        doc = getattr(self, "handle_" + fullarg).__doc__
                        print(textwrap.dedent(doc).strip())

                if i != len(args) - 1:
                    print()

    def handle_info(self, args: List[str]) -> None:
        """
        info <arg>...
          Print information about the current state of the program. Valid arguments to
          info are "registers", "stack", "flags" and "symbols". Arguments may be
          abbreviated with a unique prefix. The argument list defaults to "registers",
          "flags", and "stack" if not provided.
        """
        if args:
            try:
                fullargs = [self.expand_info_arg(arg) for arg in args]
            except HERAError as e:
                print("Error: " + str(e) + ".")
                return
        else:
            fullargs = ["registers", "flags", "stack"]

        for i, fullarg in enumerate(fullargs):
            if fullarg == "stack":
                self.info_stack()
            elif fullarg == "symbols":
                self.info_symbols()
            elif fullarg == "registers":
                self.info_registers()
            elif fullarg == "flags":
                self.info_flags()
            else:
                raise RuntimeError("this should never happen!")

            if i != len(fullargs) - 1:
                print()

    def expand_info_arg(self, arg: str) -> str:
        arg = arg.lower()
        if "stack".startswith(arg):
            # "stack" comes before "symbols" because "s" should resolve to "stack".
            return "stack"
        elif "symbols".startswith(arg):
            return "symbols"
        elif "registers".startswith(arg):
            return "registers"
        elif "flags".startswith(arg):
            return "flags"
        else:
            raise HERAError("unrecognized argument `{}`".format(arg))

    def handle_list(self, args: List[str]) -> None:
        """
        list
          Print the current line of source and the three previous and next lines.

        list <n>
          Print the current line of source code and the `n` previous and next lines.
        """
        if len(args) > 1:
            print("list takes zero or one arguments.")
            return

        try:
            context = int(args[0], base=0) if args else 3
        except ValueError:
            print("Could not parse argument to list.")
            return

        if not self.debugger.finished():
            loc = self.debugger.op().loc
            self.print_range_of_ops(loc, context=context)
        else:
            print("Program has finished executing.")

    def handle_ll(self, args: List[str]) -> None:
        """
        ll
          Print every line of the current file's source code.
        """
        if len(args) != 0:
            print("ll takes no arguments.")
            return

        if not self.debugger.finished():
            loc = self.debugger.op().loc
            self.print_range_of_ops(loc)
        else:
            print("Program has finished executing.")

    @mutates
    def handle_next(self, args: List[str]) -> None:
        """
        next
          Execute the current line. If the current line is a CALL instruction, the
          debugger executes the entire function (including nested and recursive calls)
          and moves on to the next line. If you wish to neter over the function call,
          use `step` instead.

        next <n>
          Execute the next n instructions. This command will follow branches, so be
          careful!
        """
        if len(args) > 1:
            print("next takes zero or one arguments.")
            return

        if not self.debugger.finished():
            try:
                n = int(args[0]) if args else 1
            except ValueError:
                print("Could not parse argument to next.")
                return

            for _ in range(n):
                if self.debugger.finished():
                    break
                self.debugger.next(step=False)

        self.print_current_op()

    @mutates
    def handle_off(self, args: List[str]) -> None:
        """
        off <f1> <f2>...
          Turn off all the HERA machine flags listed. Flags may be given in long
          form (carry-block, carry, overflow, sign, zero) or short form (cb, c, v,
          s, z).
        """
        if len(args) == 0:
            print("off takes one or more arguments.")
            return

        try:
            flags = resolve_flags(args)
        except HERAError as e:
            print(e)
            return

        for flag in flags:
            setattr(self.debugger.vm, flag, False)

    @mutates
    def handle_on(self, args: List[str]) -> None:
        """
        on <f1> <f2>...
          Turn on all the HERA machine flags listed. Flags may be given in long form
          (carry-block, carry, overflow, sign, zero) or short form (cb, c, v, s, z).
        """
        if len(args) == 0:
            print("on takes one or more arguments.")
            return

        try:
            flags = resolve_flags(args)
        except HERAError as e:
            print(e)
            return

        for flag in flags:
            setattr(self.debugger.vm, flag, True)

    def handle_print(self, argstr: str) -> None:
        """
        print <x> <y> <z>...
          Print the values of all the supplied arguments. The first argument may
          optionally be a format specifier, e.g. ":xds". Each character of the string
          identifies a format in which to print the value. The following formats are
          recognized: d for decimal, x for hexadecimal, o for octal, b for binary, c
          for character literals, s for signed integers, and l for source code
          locations. When not provided, the format specifier defaults to ":xdsc".

          Examples:
            A register:        print R7
            A memory location: print @1000
            A symbol:          print some_label
            Multiple values:   print R1, R2, R3
            Formatted:         print :bl PC_ret
            Arithmetic:        print @(@(FP+1)) * 7
        """
        if not argstr:
            print("print takes one or more arguments.")
            return

        try:
            tree = miniparser.parse(argstr)
        except SyntaxError as e:
            msg = str(e)
            if msg:
                print("Parse error: {}.".format(msg))
            else:
                print("Parse error.".format(msg))
        else:
            spec = tree.fmt
            for c in spec:
                if c not in "dxobcsl":
                    print("Unknown format specifier `{}`.".format(c))
                    return

                # 'c' and 's' do not always generate output, if the given value is not a
                # char or signed integer, respectively. Output can be forced with the
                # 'C' and 'S', which we do if the user explicitly provided these
                # formats.
                spec = spec.replace("c", "C")
                spec = spec.replace("s", "S")

            try:
                if len(tree.seq) > 1:
                    for arg in tree.seq:
                        self.print_one_expr(arg, spec, with_lhs=True)
                else:
                    self.print_one_expr(tree.seq[0], spec)
            except HERAError as e:
                print("Eval error: {}.".format(e))

    def print_one_expr(self, tree, spec: str, *, with_lhs=False) -> None:
        """Print a single expression with the given format specification."""

        # Customize the format specifier depending on the type of expression.
        if isinstance(tree, RegisterNode):
            # R13 is used to hold the return value of the PC in function calls,
            # so printing the location is useful.
            if tree.value == 13 and not spec:
                spec = augment_spec(spec, "l")
        elif isinstance(tree, SymbolNode):
            if tree.value.lower() == "pc":
                spec = augment_spec(spec, "l")
            else:
                try:
                    value = self.debugger.symbol_table[tree.value]
                except KeyError:
                    raise HERAError("{} is not defined".format(tree.value))
                else:
                    if isinstance(value, Label):
                        spec = augment_spec(spec, "l")
        elif isinstance(tree, IntNode):
            if not spec:
                spec = "d"

        value = self.evaluate_node(tree)
        if with_lhs:
            print("{} = {}".format(tree, self.format_int(value, spec)))
        else:
            print(self.format_int(value, spec))

    @mutates
    def handle_restart(self, args: List[str]) -> None:
        """
        restart
          Restart execution of the program from the beginning. All registers and
          memory cells are reset.
        """
        if len(args) != 0:
            print("restart takes no arguments.")
            return

        self.debugger.reset()
        self.print_current_op()

    @mutates
    def handle_step(self, args: List[str]) -> None:
        """
        step
          Step into the execution of a function.  The step command is only valid when
          the current instruction is CALL.
        """
        if len(args) > 0:
            print("step takes no arguments.")
            return

        if self.debugger.op().name != "CALL":
            print("step is only valid when the current instruction is CALL.")
            return

        self.debugger.next(step=True)
        self.print_current_op()

    def handle_undo(self, args: List[str]) -> None:
        """
        undo
          Undo the last operation that changed the state of the debugger.
        """
        if len(args) > 0:
            print("undo takes no arguments.")
            return

        if self.debugger.old is None:
            print("Nothing to undo.")
            return
        else:
            print("Undid {}.".format(self.command_history.pop()))

        self.debugger = self.debugger.old

    def info_registers(self) -> None:
        nonzero = 0
        for i, r in enumerate(self.debugger.vm.registers[1:], start=1):
            if r != 0:
                nonzero += 1
                end = ", " if i != 15 or nonzero < 15 else ""
                print("R{} = {}".format(i, r), end=end)

        if nonzero == 0:
            print("All registers set to zero.")
        elif nonzero != 15:
            print("all other registers set to zero.")
        else:
            print()

    def info_flags(self) -> None:
        vm = self.debugger.vm
        flags = []
        if vm.flag_carry_block:
            flags.append("carry-block flag is on")
        if vm.flag_carry:
            flags.append("carry flag is on")
        if vm.flag_overflow:
            flags.append("overflow flag is on")
        if vm.flag_zero:
            flags.append("zero flag is on")
        if vm.flag_sign:
            flags.append("sign flag is on")

        if len(flags) == 5:
            print("All flags are on.")
        elif len(flags) == 0:
            print("All flags are off.")
        else:
            flagstr = ", ".join(flags)
            flagstr = flagstr[0].upper() + flagstr[1:]
            print(flagstr + ", all other flags are off.")

    def info_stack(self) -> None:
        vm = self.debugger.vm
        if vm.expected_returns:
            print("Call stack (last call at bottom)")
            for call_address, return_address in vm.expected_returns:
                fname = self.debugger.find_label(call_address)
                floc = self.debugger.get_breakpoint_name(
                    call_address, append_label=False
                )
                rloc = self.debugger.get_breakpoint_name(
                    return_address - 1, append_label=False
                )
                if fname is not None:
                    print("  {} ({}, called from {})".format(fname, floc, rloc))
                else:
                    print("  {} (called from {})".format(floc, rloc))
        else:
            print("The call stack is empty.")

    def info_symbols(self) -> None:
        constants = []
        labels = []
        dlabels = []
        for key, val in self.debugger.symbol_table.items():
            if isinstance(val, Label):
                debug_info = self.debugger.program.debug_info
                labels.append("{} ({})".format(key, debug_info.labels[key]))
            elif isinstance(val, DataLabel):
                dlabels.append("{} (0x{:x})".format(key, val))
            else:
                constants.append("{} ({})".format(key, val))

        if constants:
            print("Constants: " + ", ".join(constants))

        if labels:
            print("Labels: " + ", ".join(labels))

        if dlabels:
            print("Data labels: " + ", ".join(dlabels))

    def evaluate_node(self, node) -> int:
        vm = self.debugger.vm
        if isinstance(node, IntNode):
            if node.value >= 2 ** 16 or node.value < -2 ** 15:
                raise HERAError("integer literal exceeds 16 bits")
            return node.value
        elif isinstance(node, RegisterNode):
            return vm.load_register(node.value)
        elif isinstance(node, MemoryNode):
            address = self.evaluate_node(node.address)
            return vm.load_memory(address)
        elif isinstance(node, SymbolNode):
            if node.value.lower() == "pc":
                return vm.pc
            else:
                try:
                    return self.debugger.symbol_table[node.value]
                except KeyError:
                    raise HERAError("{} is not defined".format(node.value))
        elif isinstance(node, PrefixNode):
            arg = self.evaluate_node(node.arg)
            if node.op == "-":
                result = -arg
            else:
                raise RuntimeError("unhandled prefix operator: " + node.op)

            if overflow(result):
                raise HERAError("overflow from unary " + node.op)

            return result
        elif isinstance(node, InfixNode):
            left = self.evaluate_node(node.left)
            right = self.evaluate_node(node.right)
            if node.op == "+":
                result = left + right
            elif node.op == "-":
                result = left - right
            elif node.op == "*":
                result = left * right
            elif node.op == "/":
                if right == 0:
                    raise HERAError("division by zero")
                result = left // right
            else:
                raise RuntimeError("unhandled infix operator: " + node.op)

            if overflow(result):
                raise HERAError("overflow from " + node.op)

            return result
        else:
            raise RuntimeError("unknown node type {}".format(node.__class__.__name__))

    def print_current_op(self) -> None:
        """
        Print the next operation to be executed. If the program has finished executed,
        nothing is printed.
        """
        if not self.debugger.finished():
            loc = self.debugger.op().loc
            self.print_range_of_ops(loc, context=1)
        else:
            print("Program has finished executing.")

    def print_range_of_ops(self, loc: Location, context: Optional[int] = None) -> None:
        """
        Print the line indicated by `loc`, as well as `context` previous and following
        lines. If `context` is None, the whole file is printed.
        """
        lineno = loc.line - 1
        lines = loc.file_lines
        max_lineno_width = len(str(len(lines)))

        if context is None:
            lo = 0
            hi = len(lines)
        else:
            lo = max(lineno - context, 0)
            hi = min(lineno + context + 1, len(lines))

        print("[{}]\n".format(loc.path))
        for i in range(lo, hi):
            prefix = "->  " if i == lineno else "    "
            print(prefix, end="")
            print(pad(str(i + 1), max_lineno_width), end="")

            line = lines[i].rstrip()
            if line:
                print("  " + line)
            else:
                print()

    def format_int(self, v: int, spec: str) -> str:
        if not spec:
            spec = DEFAULT_SPEC

        if "l" in spec:
            spec = spec.replace("l", "")
            loc = True
        else:
            loc = False

        if loc:
            try:
                label = self.debugger.get_breakpoint_name(v, append_label=False)
            except IndexError:
                return format_int(v, spec=spec)
            else:
                return format_int(v, spec=spec) + " [" + label + "]"
        else:
            return format_int(v, spec=spec)

    def asm_settings(self, *, code, data):
        """Override some defaults in self.settings for the assembler."""
        settings = copy.copy(self.settings)
        settings.stdout = True
        settings.code = code
        settings.data = data
        return settings


DEFAULT_SPEC = "dsc"


def augment_spec(spec: str, f: str) -> str:
    """Augment the format specifier with the additional format character."""
    if not spec:
        return augment_spec(DEFAULT_SPEC, f)
    else:
        return spec + f if f not in spec else spec


FLAG_SHORT_TO_LONG = {
    "cb": "carry_block",
    "c": "carry",
    "v": "overflow",
    "s": "sign",
    "z": "zero",
}


def resolve_flags(args: List[str]) -> List[str]:
    flags = []
    for arg in args:
        arg = arg.replace("-", "_")
        if arg not in FLAG_SHORT_TO_LONG.values():
            try:
                longflag = FLAG_SHORT_TO_LONG[arg]
            except KeyError:
                raise HERAError("Unrecognized flag: `{}`.".format(arg))
            else:
                flags.append("flag_" + longflag)
        else:
            flags.append("flag_" + arg)
    return flags


def overflow(v: int) -> bool:
    return v >= 2 ** 16 or v < -2 ** 15


HELP = """\
Available commands:
    asm <op>        Show the binary machine code that the HERA operations
                    assembles to.

    assign <x> <y>  Assign the value of y to x.

    break <loc>     Set a breakpoint at the given location. When no arguments
                    are given, all current breakpoints are printed.

    clear <loc>     Clear a breakpoint at the given location.

    continue        Execute the program until a breakpoint is encountered or
                    the program terminates.

    dis <n>         Disassemble the 16-bit integer into a HERA operation.

    execute <op>    Execute a HERA operation.

    goto <loc>      Jump to the given location.

    help            Print this help message.

    info            Print information about the current state of the program.

    list <n>        Print the current lines of source code and the n previous
                    and next lines. If not provided, n defaults to 3.

    ll              Print the entire program.

    next            Execute the current line.

    off <flag>      Turn the given machine flag off.

    on <flag>       Turn the given machine flag on.

    print <x>       Print the value of x.

    restart         Restart the execution of the program from the beginning.

    step            Step into the execution of a function.

    undo            Undo the last operation.

    quit            Exit the debugger.

    <x> = <y>       Alias for "assign <x> <y>".

Command names can generally be abbreviated with a unique prefix, e.g. "n" for
"next".
"""



from .utils import execute_program_helper


def test_warning_for_stack_overflow(capsys):
    program = "SET(R1, 0xC002)\nADD(SP, SP, R1)\nDEC(SP, 10)\nINC(SP, 30)"
    execute_program_helper(program)

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Warning: stack has overflowed into data segment, line 2 col 1 of <stdin>

  ADD(SP, SP, R1)
  ^


Virtual machine state after execution:
    R1  = 0xc002 = 49154 = -16382

    Carry-block flag is OFF
    Carry flag is OFF
    Overflow flag is OFF
    Zero flag is OFF
    Sign flag is ON

1 warning emitted.
"""
    )


def test_warning_for_zero_prefixed_octal(capsys):
    execute_program_helper("SET(R1, 016)\nSET(R2, 017)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Warning: consider using "0o" prefix for octal numbers, line 1 col 9 of <stdin>

  SET(R1, 016)
          ^

Warning: consider using "0o" prefix for octal numbers, line 2 col 9 of <stdin>

  SET(R2, 017)
          ^


Virtual machine state after execution:
    R1  = 0x000e = 14
    R2  = 0x000f = 15

    All flags are OFF

2 warnings emitted.
"""
    )


def test_no_warning_for_zero_prefixed_octal_with_flag(capsys):
    execute_program_helper("SET(R1, 016)\nSET(R2, 017)", flags=["--warn-octal-off"])

    captured = capsys.readouterr().err
    assert "Warning" not in captured


def test_warning_for_invalid_backslash_escape_in_character(capsys):
    execute_program_helper("SET(R1, '\\u')")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Warning: unrecognized backslash escape, line 1 col 11 of <stdin>

  SET(R1, '\\u')
            ^


Virtual machine state after execution:
    R1  = 0x0075 = 117 = 'u'

    All flags are OFF

1 warning emitted.
"""
    )


def test_warning_for_invalid_backslash_escape_in_string(capsys):
    execute_program_helper('LP_STRING("bad: \\o")')

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Warning: unrecognized backslash escape, line 1 col 18 of <stdin>

  LP_STRING("bad: \\o")
                   ^


Virtual machine state after execution:
    R1 through R10 are all zero.

    All flags are OFF

1 warning emitted.
"""
    )


def test_warning_for_R11_with_NOT(capsys):
    execute_program_helper("SET(R11, 5)\nNOT(R1, R11)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Warning: don't use R11 with NOT, line 2 col 9 of <stdin>

  NOT(R1, R11)
          ^


Virtual machine state after execution:
    R1 through R10 are all zero.

    Carry-block flag is OFF
    Carry flag is OFF
    Overflow flag is OFF
    Zero flag is ON
    Sign flag is OFF

1 warning emitted.
"""
    )


def test_warning_for_improper_register_with_CALL(capsys):
    execute_program_helper("CALL(R11, l)\nLABEL(l)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Warning: first argument to CALL should be R12, line 1 col 6 of <stdin>

  CALL(R11, l)
       ^


Virtual machine state after execution:
    R1 through R10 are all zero.

    All flags are OFF

1 warning emitted.
"""
    )


def test_warning_for_improper_first_register_with_RETURN(capsys):
    execute_program_helper("RETURN(R11, R13)", flags=["--warn-return-off"])

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Warning: first argument to RETURN should be R12, line 1 col 8 of <stdin>

  RETURN(R11, R13)
         ^


Virtual machine state after execution:
    R1 through R10 are all zero.

    All flags are OFF

1 warning emitted.
"""
    )


def test_warning_for_improper_second_register_with_RETURN(capsys):
    execute_program_helper("RETURN(R12, R11)", flags=["--warn-return-off"])

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Warning: second argument to RETURN should be R13, line 1 col 13 of <stdin>

  RETURN(R12, R11)
              ^


Virtual machine state after execution:
    R1 through R10 are all zero.

    All flags are OFF

1 warning emitted.
"""
    )


def test_warning_for_unnecessary_include(capsys):
    execute_program_helper("#include <HERA.h>")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Warning: #include <HERA.h> is not necessary for hera-py, line 1 col 11 of <stdin>

  #include <HERA.h>
            ^


Virtual machine state after execution:
    R1 through R10 are all zero.

    All flags are OFF

1 warning emitted.
"""
    )


def test_warning_for_bad_RETURN_value(capsys):
    program = """\
CALL(FP_alt, do_nothing)
HALT()

LABEL(do_nothing)
  SET(PC_ret, 100)
  RETURN(FP_alt, PC_ret)
    """
    execute_program_helper(program)

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Warning: incorrect return address (got 100, expected 3), line 6 col 3 of <stdin>

    RETURN(FP_alt, PC_ret)
    ^


Virtual machine state after execution:
    R1 through R10 are all zero.

    All flags are OFF

1 warning emitted.
"""
    )



from hera.checker import convert_ops
from hera.data import Token
from hera.op import ADD, BR, BRR, CALL, FOFF, FON, INTEGER, SET, SETHI, SETLO, SUB
from hera.parser import parse


def helper(argstr):
    oplist, messages = parse(argstr)
    assert not messages.errors
    return oplist[0].convert()


def test_convert_register_branch_with_register():
    ops = helper("BR(R1)")
    assert ops == [BR(Token.R(1))]


def test_convert_register_branch_with_integer():
    # Simulates BR(l) where l = 1000
    ops = helper("BR(1000)")
    assert ops == [
        SETLO(Token.R(11), Token.Int(232)),
        SETHI(Token.R(11), Token.Int(3)),
        BR(Token.R(11)),
    ]


def test_convert_SET():
    ops = helper("SET(R1, 0b100010101010)")
    assert ops == [
        SETLO(Token.R(1), Token.Int(0b10101010)),
        SETHI(Token.R(1), Token.Int(0b1000)),
    ]


def test_convert_CALL():
    # Simulates CALL(R12, l) where l = 1000
    ops = helper("CALL(R12, 1000)")
    assert ops == [
        SETLO(Token.R(13), Token.Int(232)),
        SETHI(Token.R(13), Token.Int(3)),
        CALL(Token.R(12), Token.R(13)),
    ]


def test_convert_CMP():
    ops = helper("CMP(R5, R6)")
    assert ops == [FON(Token.Int(8)), SUB(Token.R(0), Token.R(5), Token.R(6))]


def test_convert_CON():
    ops = helper("CON()")
    assert ops == [FON(Token.Int(8))]


def test_convert_COFF():
    ops = helper("COFF()")
    assert ops == [FOFF(Token.Int(8))]


def test_convert_CBON():
    ops = helper("CBON()")
    assert ops == [FON(Token.Int(0x10))]


def test_convert_CCBOFF():
    ops = helper("CCBOFF()")
    assert ops == [FOFF(Token.Int(0x18))]


def test_convert_FLAGS():
    ops = helper("FLAGS(R14)")
    assert ops == [FOFF(Token.Int(8)), ADD(Token.R(0), Token.R(14), Token.R(0))]


def test_convert_LABEL():
    ops = helper("LABEL(l)")
    assert ops == []


def test_convert_DLABEL():
    ops = helper("DLABEL(l)")
    assert ops == []


def test_convert_NOP():
    ops = helper("NOP()")
    assert ops == [BRR(Token.Int(1))]


def test_convert_HALT():
    ops = helper("HALT()")
    assert ops == [BRR(Token.Int(0))]


def test_convert_non_pseudo_op():
    ops = helper("ADD(R1, R2, R3)")
    assert ops == [ADD(Token.R(1), Token.R(2), Token.R(3))]


def test_convert_ops_with_constant():
    oplist, messages = convert_ops([SET(Token.R(1), Token.Sym("n"))], {"n": 100})

    assert len(messages.errors) == 0
    assert oplist == [
        SETLO(Token.R(1), Token.Int(100)),
        SETHI(Token.R(1), Token.Int(0)),
    ]


def test_convert_ops_reports_error_for_invalid_relative_branch():
    oplist, messages = convert_ops([BRR(Token.Sym("l"))], {"l": 9000})

    assert len(messages.errors) == 1
    assert "too far" in messages.errors[0][0]


def test_convert_ops_does_not_report_error_for_valid_relative_branch():
    data = [INTEGER(Token.Int(1))] * 200
    oplist, messages = convert_ops(data + [BRR(Token.Sym("l"))], {"l": 1})

    assert len(messages.errors) == 0
    assert oplist == data + [BRR(Token.Int(1))]



"""
The definition of all the operations in the HERA language.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: March 2019
"""
import json
import sys
from contextlib import suppress
from typing import Dict, List, Optional, Union

from hera import stdlib
from hera.data import Constant, DataLabel, HERAError, Label, Location, Messages, Token
from hera.utils import format_int, from_u16, print_error, print_warning, to_u16, to_u32
from hera.vm import VirtualMachine


class AbstractOperation:
    """
    The abstract base class for HERA operations. All operation classes should inherit
    from this class.
    """

    BITV = ""

    def __init__(self, *args, loc=None):
        self.args = [a.value for a in args]
        self.tokens = list(args)
        # When the preprocessor converts pseudo-ops to real ops, it will set the
        # original field of the real op to the corresponding pseudo-op, for use by the
        # HERA debugger.
        self.original = None

        if isinstance(loc, Location):
            self.loc = loc
        elif hasattr(loc, "location"):
            self.loc = loc.location
        else:
            self.loc = None

    def typecheck(self, symbol_table: Dict[str, int]) -> Messages:
        """
        Type-check the operation. Subclasses do not generally need to override this
        method, as long as they provide a P class field listing their parameter types.
        """
        messages = Messages()
        if len(self.P) < len(self.tokens):
            msg = "too many args to {} (expected {})".format(self.name, len(self.P))
            messages.err(msg, self.loc)
        elif len(self.P) > len(self.tokens):
            msg = "too few args to {} (expected {})".format(self.name, len(self.P))
            messages.err(msg, self.loc)

        return messages.extend(check_arglist(self.P, self.tokens, symbol_table))

    def convert(self) -> List["AbstractOperation"]:
        """
        Convert the pseudo-operation into a list of real operations. Only pseudo-ops
        need to override this method.
        """
        return [self]

    def assemble(self) -> bytes:
        """
        Assemble the operation into a 16-bit string. Subclasses do not generally need
        to override this method, as long as they provide a BITV class field.
        """
        return substitute_bitvector(self.BITV, self.args)

    @classmethod
    def disassemble(cls, *args):
        """
        Disassemble the integer arguments into an operation. Subclasses do not
        generally need to override this method; it is provided only for the special
        purposes of the INC and DEC classes.
        """
        return cls(*args)

    def execute(self, vm: VirtualMachine) -> None:
        """
        Execute the operation on a virtual machine. Real operations should override
        this method; pseudo-operations should not.
        """
        raise NotImplementedError

    def __getattr__(self, name):
        if name == "name":
            return self.__class__.__name__
        else:
            raise AttributeError(name)

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and len(self.tokens) == len(other.tokens)
            and all(a1 == a2 for a1, a2 in zip(self.tokens, other.tokens))
        )

    def __repr__(self):
        return "{}({})".format(self.name, ", ".join(repr(a) for a in self.tokens))

    def __str__(self):
        return "{}({})".format(
            self.name, ", ".join(arg_to_string(a) for a in self.tokens)
        )


def arg_to_string(arg):
    if arg.type == Token.STRING:
        return json.dumps(arg.value)
    elif arg.type == Token.REGISTER:
        return "R" + str(arg.value)
    else:
        return str(arg.value)


REGISTER = "REGISTER"
REGISTER_OR_LABEL = "REGISTER_OR_LABEL"
STRING = "STRING"
LABEL_TYPE = "LABEL_TYPE"
I16 = range(-2 ** 15, 2 ** 16)
I16_OR_LABEL = "I16_OR_LABEL"
U16 = range(2 ** 16)
I8 = range(-2 ** 7, 2 ** 8)
I8_OR_LABEL = "I8_OR_LABEL"
U5 = range(2 ** 5)
U4 = range(2 ** 4)


class UnaryOp(AbstractOperation):
    """
    Abstract class to simplify implementation of unary operations. Child classes only
    need to implement the calculate method and set the BITV field.
    """

    P = (REGISTER, REGISTER)

    def execute(self, vm):
        arg = vm.load_register(self.args[1])
        result = self.calculate(vm, arg)
        vm.set_zero_and_sign(result)
        vm.store_register(self.args[0], result)
        vm.pc += 1

    @staticmethod
    def calculate(vm, arg):
        """
        Calculate the result of the unary operation, given a concrete argument.

        The virtual machine is passed in so that the overflow and carry flags can be
        set if necessary.
        """
        raise NotImplementedError


class BinaryOp(AbstractOperation):
    """
    Abstract class to simplify implementation of binary operations. Child classes only
    need to implement the calculate method and set the BITV field.
    """

    P = (REGISTER, REGISTER, REGISTER)

    def execute(self, vm):
        left = vm.load_register(self.args[1])
        right = vm.load_register(self.args[2])
        result = self.calculate(vm, left, right)
        vm.set_zero_and_sign(result)
        vm.store_register(self.args[0], result)
        vm.pc += 1

    @staticmethod
    def calculate(vm, left, right):
        """
        Calculate the result of the binary operation, given the concrete left and right
        arguments.

        The virtual machine is passed in so that the overflow and carry flags can be
        set if necessary.
        """
        raise NotImplementedError


class Branch(AbstractOperation):
    pass


class RegisterBranch(Branch):
    """
    Abstract class to simplify implementation of register branches. Child classes only
    need to implement the should method and set the BITV field.
    """

    P = (REGISTER_OR_LABEL,)

    def execute(self, vm):
        if self.should(vm):
            vm.pc = vm.load_register(self.args[0])
        else:
            vm.pc += 1

    @staticmethod
    def should(vm):
        """
        Return True if branching should occur, based on the virtual machine's state.
        """
        raise NotImplementedError

    def convert(self):
        if self.tokens[0].type == Token.REGISTER:
            # When the argument to the branch is a concrete register.
            return super().convert()
        else:
            # When the argument to the branch is a label, which has already been
            # substituted for its value.
            lbl = self.args[0]
            return [
                SETLO(Token.R(11), Token.Int(lbl & 0xFF)),
                SETHI(Token.R(11), Token.Int(lbl >> 8)),
                self.__class__(Token.R(11)),
            ]


class RelativeBranch(Branch):
    """
    Abstract class to simplify implementation of relative branches. Child classes only
    need to implement the should method and set the BITV field.
    """

    P = (I8_OR_LABEL,)

    def execute(self, vm):
        if self.should(vm):
            vm.pc += self.args[0]
        else:
            vm.pc += 1

    @staticmethod
    def should(vm):
        """
        Return True if branching should occur, based on the virtual machine's state.
        """
        raise NotImplementedError


class DebuggingOperation(AbstractOperation):
    pass


class DataOperation(AbstractOperation):
    pass


class SETLO(AbstractOperation):
    P = (REGISTER, I8)
    BITV = "1110 AAAA bbbbbbbb"

    def execute(self, vm):
        value = self.args[1]
        if value > 127:
            value -= 256

        vm.store_register(self.args[0], to_u16(value))
        vm.pc += 1


class SETHI(AbstractOperation):
    P = (REGISTER, I8)
    BITV = "1111 AAAA bbbbbbbb"

    def execute(self, vm):
        target, value = self.args
        vm.store_register(target, (value << 8) + (vm.load_register(target) & 0x00FF))
        vm.pc += 1


class SET(AbstractOperation):
    P = (REGISTER, I16_OR_LABEL)

    def convert(self):
        dest = self.tokens[0]
        value = to_u16(self.args[1])
        lo = value & 0xFF
        hi = value >> 8
        return [
            SETLO(dest, Token.Int(lo), loc=self.loc),
            SETHI(dest, Token.Int(hi), loc=self.loc),
        ]


class ADD(BinaryOp):
    BITV = "1010 AAAA BBBB CCCC"

    @staticmethod
    def calculate(vm, left, right):
        carry = 1 if not vm.flag_carry_block and vm.flag_carry else 0

        result = (left + right + carry) & 0xFFFF

        vm.flag_carry = result < (left + right + carry)
        vm.flag_overflow = from_u16(result) != from_u16(left) + from_u16(right)

        return result


class SUB(BinaryOp):
    BITV = "1011 AAAA BBBB CCCC"

    @staticmethod
    def calculate(vm, left, right):
        borrow = 1 if not vm.flag_carry_block and not vm.flag_carry else 0

        # to_u16 is necessary because although left and right are necessarily
        # uints, left - right - borrow might not be.
        result = to_u16((left - right - borrow) & 0xFFFF)

        vm.flag_carry = left >= right
        vm.flag_overflow = from_u16(result) != from_u16(left) - from_u16(right) - borrow

        return result


class MUL(BinaryOp):
    BITV = "1100 AAAA BBBB CCCC"

    @staticmethod
    def calculate(vm, left, right):
        if vm.flag_sign and not vm.flag_carry_block:
            # Take the high 16 bits.
            left = to_u32(from_u16(left))
            right = to_u32(from_u16(right))
            result = ((left * right) & 0xFFFF0000) >> 16
        else:
            # Take the low 16 bits.
            result = (left * right) & 0xFFFF

        vm.flag_carry = result < left * right
        vm.flag_overflow = from_u16(result) != from_u16(left) * from_u16(right)

        return result


class AND(BinaryOp):
    BITV = "1000 AAAA BBBB CCCC"

    @staticmethod
    def calculate(vm, left, right):
        return left & right


class OR(BinaryOp):
    BITV = "1001 AAAA BBBB CCCC"

    @staticmethod
    def calculate(vm, left, right):
        return left | right


class XOR(BinaryOp):
    BITV = "1101 AAAA BBBB CCCC"

    @staticmethod
    def calculate(vm, left, right):
        return left ^ right


class INC(AbstractOperation):
    P = (REGISTER, range(1, 65))
    BITV = "0011 AAAA 10bb bbbb"

    def execute(self, vm):
        target, value = self.args

        original = vm.load_register(target)
        result = (value + original) & 0xFFFF
        vm.store_register(target, result)

        vm.set_zero_and_sign(result)
        vm.flag_overflow = from_u16(result) != from_u16(original) + value
        vm.flag_carry = value + original >= 2 ** 16
        vm.pc += 1

    def assemble(self):
        # The increment value encoded in the instruction is one less than the actual
        # increment, i.e. INC(R1, 1) is assembled as if it were INC(R1, 0) since the
        # latter is illegal.
        bv = super().assemble()
        return bytes([bv[0], bv[1] - 1])

    @classmethod
    def disassemble(cls, arg0, arg1):
        return cls(arg0, Token(Token.INT, arg1.value + 1))


class DEC(AbstractOperation):
    P = (REGISTER, range(1, 65))
    BITV = "0011 AAAA 11bb bbbb"

    def execute(self, vm):
        target, value = self.args

        original = vm.load_register(target)
        result = to_u16((original - value) & 0xFFFF)
        vm.store_register(target, result)

        vm.set_zero_and_sign(result)
        vm.flag_overflow = from_u16(result) != from_u16(original) - value
        vm.flag_carry = original < value
        vm.pc += 1

    def assemble(self):
        # The decrement value encoded in the instruction is one less than the actual
        # decrement, i.e. DEC(R1, 1) is assembled as if it were DEC(R1, 0) since the
        # latter is illegal.
        bv = super().assemble()
        return bytes([bv[0], bv[1] - 1])

    @classmethod
    def disassemble(cls, arg0, arg1):
        return cls(arg0, Token(Token.INT, arg1.value + 1))


class LSL(UnaryOp):
    BITV = "0011 AAAA 0000 BBBB"

    @staticmethod
    def calculate(vm, arg):
        carry = 1 if vm.flag_carry and not vm.flag_carry_block else 0
        result = ((arg << 1) + carry) & 0xFFFF

        vm.flag_carry = arg & 0x8000

        return result


class LSR(UnaryOp):
    BITV = "0011 AAAA 0001 BBBB"

    @staticmethod
    def calculate(vm, arg):
        carry = 2 ** 15 if vm.flag_carry and not vm.flag_carry_block else 0
        result = (arg >> 1) + carry

        vm.flag_carry = arg % 2 == 1

        return result


class LSL8(UnaryOp):
    BITV = "0011 AAAA 0010 BBBB"

    @staticmethod
    def calculate(vm, arg):
        return (arg << 8) & 0xFFFF


class LSR8(UnaryOp):
    BITV = "0011 AAAA 0011 BBBB"

    @staticmethod
    def calculate(vm, arg):
        return arg >> 8


class ASL(UnaryOp):
    BITV = "0011 AAAA 0100 BBBB"

    @staticmethod
    def calculate(vm, arg):
        carry = 1 if vm.flag_carry and not vm.flag_carry_block else 0
        result = ((arg << 1) + carry) & 0xFFFF

        vm.flag_carry = arg & 0x8000
        vm.flag_overflow = arg & 0x8000 and not result & 0x8000

        return result


class ASR(UnaryOp):
    BITV = "0011 AAAA 0101 BBBB"

    @staticmethod
    def calculate(vm, arg):
        # This is a little messy because right shift in Python rounds towards
        # negative infinity (7 >> 1 == -4) but in HERA it rounds towards zero
        # (7 >> 1 == -3).
        if arg & 0x8000:
            if arg & 0x0001:
                result = ((arg >> 1) | 0x8000) + 1
            else:
                result = arg >> 1 | 0x8000
        else:
            result = arg >> 1

        vm.flag_carry = arg & 0x0001

        return result


class SAVEF(AbstractOperation):
    P = (REGISTER,)
    BITV = "0011 AAAA 0111 0000"

    def execute(self, vm):
        value = (
            int(vm.flag_sign)
            + 2 * int(vm.flag_zero)
            + 4 * int(vm.flag_overflow)
            + 8 * int(vm.flag_carry)
            + 16 * int(vm.flag_carry_block)
        )
        vm.store_register(self.args[0], value)
        vm.pc += 1


class RSTRF(AbstractOperation):
    P = (REGISTER,)
    BITV = "0011 AAAA 0111 1000"

    def execute(self, vm):
        value = vm.load_register(self.args[0])
        vm.flag_sign = bool(value & 1)
        vm.flag_zero = bool(value & 0b10)
        vm.flag_overflow = bool(value & 0b100)
        vm.flag_carry = bool(value & 0b1000)
        vm.flag_carry_block = bool(value & 0b10000)
        vm.pc += 1


class FON(AbstractOperation):
    P = (U5,)
    BITV = "0011 000a 0110 aaaa"

    def execute(self, vm):
        value = self.args[0]
        vm.flag_sign = vm.flag_sign or bool(value & 1)
        vm.flag_zero = vm.flag_zero or bool(value & 0b10)
        vm.flag_overflow = vm.flag_overflow or bool(value & 0b100)
        vm.flag_carry = vm.flag_carry or bool(value & 0b1000)
        vm.flag_carry_block = vm.flag_carry_block or bool(value & 0b10000)
        vm.pc += 1


class FOFF(AbstractOperation):
    P = (U5,)
    BITV = "0011 100a 0110 aaaa"

    def execute(self, vm):
        value = self.args[0]
        vm.flag_sign = vm.flag_sign and not bool(value & 1)
        vm.flag_zero = vm.flag_zero and not bool(value & 0b10)
        vm.flag_overflow = vm.flag_overflow and not bool(value & 0b100)
        vm.flag_carry = vm.flag_carry and not bool(value & 0b1000)
        vm.flag_carry_block = vm.flag_carry_block and not bool(value & 0b10000)
        vm.pc += 1


class FSET5(AbstractOperation):
    P = (U5,)
    BITV = "0011 010a 0110 aaaa"

    def execute(self, vm):
        value = self.args[0]
        vm.flag_sign = bool(value & 1)
        vm.flag_zero = bool(value & 0b10)
        vm.flag_overflow = bool(value & 0b100)
        vm.flag_carry = bool(value & 0b1000)
        vm.flag_carry_block = bool(value & 0b10000)
        vm.pc += 1


class FSET4(AbstractOperation):
    P = (U4,)
    BITV = "0011 110a 0110 aaaa"

    def execute(self, vm):
        value = self.args[0]
        vm.flag_sign = bool(value & 1)
        vm.flag_zero = bool(value & 0b10)
        vm.flag_overflow = bool(value & 0b100)
        vm.flag_carry = bool(value & 0b1000)
        vm.pc += 1


class LOAD(AbstractOperation):
    P = (REGISTER, U5, REGISTER)
    BITV = "010b AAAA bbbb CCCC"

    def execute(self, vm):
        target, offset, address = self.args

        result = vm.load_memory(vm.load_register(address) + offset)
        vm.set_zero_and_sign(result)
        vm.store_register(target, result)
        vm.pc += 1


class STORE(AbstractOperation):
    P = (REGISTER, U5, REGISTER)
    BITV = "011b AAAA bbbb CCCC"

    def execute(self, vm):
        source, offset, address = self.args

        vm.store_memory(vm.load_register(address) + offset, vm.load_register(source))
        vm.pc += 1


class BR(RegisterBranch):
    BITV = "0001 0000 0000 AAAA"

    @staticmethod
    def should(vm):
        return True


class BRR(RelativeBranch):
    BITV = "0000 0000 aaaa aaaa"

    def execute(self, vm):
        if self.args[0] != 0:
            vm.pc += self.args[0]
        else:
            vm.halted = True


class BL(RegisterBranch):
    BITV = "0001 0010 0000 AAAA"

    @staticmethod
    def should(vm):
        return vm.flag_sign ^ vm.flag_overflow


class BLR(RelativeBranch):
    BITV = "0000 0010 aaaa aaaa"

    @staticmethod
    def should(vm):
        return vm.flag_sign ^ vm.flag_overflow


class BGE(RegisterBranch):
    BITV = "0001 0011 0000 AAAA"

    @staticmethod
    def should(vm):
        return not (vm.flag_sign ^ vm.flag_overflow)


class BGER(RelativeBranch):
    BITV = "0000 0011 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not (vm.flag_sign ^ vm.flag_overflow)


class BLE(RegisterBranch):
    BITV = "0001 0100 0000 AAAA"

    @staticmethod
    def should(vm):
        return (vm.flag_sign ^ vm.flag_overflow) or vm.flag_zero


class BLER(RelativeBranch):
    BITV = "0000 0100 aaaa aaaa"

    @staticmethod
    def should(vm):
        return (vm.flag_sign ^ vm.flag_overflow) or vm.flag_zero


class BG(RegisterBranch):
    BITV = "0001 0101 0000 AAAA"

    @staticmethod
    def should(vm):
        return not (vm.flag_sign ^ vm.flag_overflow) and not vm.flag_zero


class BGR(RelativeBranch):
    BITV = "0000 0101 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not (vm.flag_sign ^ vm.flag_overflow) and not vm.flag_zero


class BULE(RegisterBranch):
    BITV = "0001 0110 0000 AAAA"

    @staticmethod
    def should(vm):
        return not vm.flag_carry or vm.flag_zero


class BULER(RelativeBranch):
    BITV = "0000 0110 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not vm.flag_carry or vm.flag_zero


class BUG(RegisterBranch):
    BITV = "0001 0111 0000 AAAA"

    @staticmethod
    def should(vm):
        return vm.flag_carry and not vm.flag_zero


class BUGR(RelativeBranch):
    BITV = "0000 0111 aaaa aaaa"

    @staticmethod
    def should(vm):
        return vm.flag_carry and not vm.flag_zero


class BZ(RegisterBranch):
    BITV = "0001 1000 0000 AAAA"

    @staticmethod
    def should(vm):
        return vm.flag_zero


class BZR(RelativeBranch):
    BITV = "0000 1000 aaaa aaaa"

    @staticmethod
    def should(vm):
        return vm.flag_zero


class BNZ(RegisterBranch):
    BITV = "0001 1001 0000 AAAA"

    @staticmethod
    def should(vm):
        return not vm.flag_zero


class BNZR(RelativeBranch):
    BITV = "0000 1001 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not vm.flag_zero


class BC(RegisterBranch):
    BITV = "0001 1010 0000 AAAA"

    @staticmethod
    def should(vm):
        return vm.flag_carry


class BCR(RelativeBranch):
    BITV = "0000 1010 aaaa aaaa"

    @staticmethod
    def should(vm):
        return vm.flag_carry


class BNC(RegisterBranch):
    BITV = "0001 1011 0000 AAAA"

    @staticmethod
    def should(vm):
        return not vm.flag_carry


class BNCR(RelativeBranch):
    BITV = "0000 1011 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not vm.flag_carry


class BS(RegisterBranch):
    BITV = "0001 1100 0000 AAAA"

    @staticmethod
    def should(vm):
        return vm.flag_sign


class BSR(RelativeBranch):
    BITV = "0000 1100 aaaa aaaa"

    @staticmethod
    def should(vm):
        return vm.flag_sign


class BNS(RegisterBranch):
    BITV = "0001 1101 0000 AAAA"

    @staticmethod
    def should(vm):
        return not vm.flag_sign


class BNSR(RelativeBranch):
    BITV = "0000 1101 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not vm.flag_sign


class BV(RegisterBranch):
    BITV = "0001 1110 0000 AAAA"

    @staticmethod
    def should(vm):
        return vm.flag_overflow


class BVR(RelativeBranch):
    BITV = "0000 1110 aaaa aaaa"

    @staticmethod
    def should(vm):
        return vm.flag_overflow


class BNV(RegisterBranch):
    BITV = "0001 1111 0000 AAAA"

    @staticmethod
    def should(vm):
        return not vm.flag_overflow


class BNVR(RelativeBranch):
    BITV = "0000 1111 aaaa aaaa"

    @staticmethod
    def should(vm):
        return not vm.flag_overflow


class CALL_AND_RETURN(Branch):
    P = (REGISTER, REGISTER_OR_LABEL)

    def execute(self, vm):
        ra, rb = self.args

        old_pc = vm.pc
        vm.pc = vm.load_register(rb)
        vm.store_register(rb, old_pc + 1)
        # FP = R14
        old_fp = vm.load_register(14)
        vm.store_register(14, vm.load_register(ra))
        vm.store_register(ra, old_fp)

    def typecheck(self, *args, **kwargs):
        messages = super().typecheck(*args, **kwargs)
        if len(self.tokens) >= 1 and self.tokens[0].type == Token.REGISTER:
            with suppress(ValueError):
                if self.tokens[0].value != 12:
                    msg = "first argument to {} should be R12".format(self.name)
                    messages.warn(msg, self.tokens[0])
        return messages


class CALL(CALL_AND_RETURN):
    P = (REGISTER, REGISTER_OR_LABEL)
    BITV = "0010 0000 AAAA BBBB"

    def convert(self):
        if self.tokens[1].type == Token.REGISTER:
            return super().convert()
        else:
            return SET(Token.R(13), self.tokens[1]).convert() + [
                self.__class__(self.tokens[0], Token.R(13))
            ]

    def execute(self, vm):
        # Push a (call_address, return_address) pair onto the debugging call stack.
        vm.expected_returns.append((vm.load_register(self.args[1]), vm.pc + 1))
        super().execute(vm)


class RETURN(CALL_AND_RETURN):
    P = (REGISTER, REGISTER)
    BITV = "0010 0001 AAAA BBBB"

    def typecheck(self, *args, **kwargs):
        messages = super().typecheck(*args, **kwargs)
        if len(self.tokens) >= 2 and self.tokens[1].type == Token.REGISTER:
            with suppress(ValueError):
                if self.tokens[1].value != 13:
                    messages.warn(
                        "second argument to RETURN should be R13", self.tokens[1]
                    )
        return messages

    def execute(self, vm):
        got = vm.load_register(self.args[1])
        if vm.settings.warn_return_on:
            if vm.expected_returns:
                _, expected = vm.expected_returns.pop()
                if expected != got:
                    msg = "incorrect return address (got {}, expected {})".format(
                        got, expected
                    )
                    print_warning(vm.settings, msg, loc=vm.location)
                    vm.settings.warning_count += 1
            else:
                msg = "incorrect return address (got {}, expected <nothing>)".format(
                    got
                )
                print_warning(vm.settings, msg, loc=vm.location)
                vm.settings.warning_count += 1
        super().execute(vm)


class SWI(AbstractOperation):
    P = (U4,)
    BITV = "0010 0010 0000 aaaa"


class RTI(AbstractOperation):
    P = ()
    BITV = "0010 0011 0000 0000"


class CMP(AbstractOperation):
    P = (REGISTER, REGISTER)

    def convert(self):
        return [FON(Token.Int(8)), SUB(Token.R(0), self.tokens[0], self.tokens[1])]


class CON(AbstractOperation):
    P = ()

    def convert(self):
        return [FON(Token.Int(8))]


class COFF(AbstractOperation):
    P = ()

    def convert(self):
        return [FOFF(Token.Int(8))]


class CBON(AbstractOperation):
    P = ()

    def convert(self):
        return [FON(Token.Int(16))]


class CCBOFF(AbstractOperation):
    P = ()

    def convert(self):
        return [FOFF(Token.Int(24))]


class MOVE(AbstractOperation):
    P = (REGISTER, REGISTER)

    def convert(self):
        return [OR(self.tokens[0], self.tokens[1], Token.R(0))]


class SETRF(AbstractOperation):
    P = (REGISTER, I16_OR_LABEL)

    def convert(self):
        return SET(*self.args).convert() + FLAGS(self.tokens[0]).convert()


class FLAGS(AbstractOperation):
    P = (REGISTER,)

    def convert(self):
        return [FOFF(Token.Int(8)), ADD(Token.R(0), self.tokens[0], Token.R(0))]


class HALT(AbstractOperation):
    P = ()

    def convert(self):
        return [BRR(Token.Int(0))]


class NOP(AbstractOperation):
    P = ()

    def convert(self):
        return [BRR(Token.Int(1))]


class NEG(AbstractOperation):
    P = (REGISTER, REGISTER)

    def convert(self):
        return [FON(Token.Int(8)), SUB(self.tokens[0], Token.R(0), self.tokens[1])]


class NOT(AbstractOperation):
    P = (REGISTER, REGISTER)

    def typecheck(self, *args, **kwargs):
        messages = super().typecheck(*args, **kwargs)
        if len(self.tokens) == 2 and self.tokens[1].type == Token.REGISTER:
            with suppress(ValueError):
                if self.tokens[1].value == 11:
                    messages.warn("don't use R11 with NOT", self.tokens[1])
        return messages

    def convert(self):
        return [
            SETLO(Token.R(11), Token.Int(0xFF)),
            SETHI(Token.R(11), Token.Int(0xFF)),
            XOR(self.tokens[0], Token.R(11), self.tokens[1]),
        ]


class OPCODE(AbstractOperation):
    P = (U16,)

    def typecheck(self, *args, **kwargs):
        messages = super().typecheck(*args, **kwargs)

        try:
            disassemble(self.args[0])
        except HERAError:
            messages.err("not a HERA instruction", self.tokens[0])

        return messages

    def convert(self):
        # We don't need to call .convert() on the return value of disassemble because
        # disassemble never returns a pseudo-op.
        return [disassemble(self.args[0])]


class INTEGER(DataOperation):
    P = (I16,)

    def execute(self, vm):
        vm.store_memory(vm.dc, to_u16(self.args[0]))
        vm.dc += 1

    def assemble(self):
        return bytes([self.args[0] & 0xFF00, self.args[0] & 0xFF])


class DSKIP(DataOperation):
    P = (U16,)

    def execute(self, vm):
        vm.dc += self.args[0]

    def assemble(self):
        return bytes([0] * self.args[0] * 2)


class LP_STRING(DataOperation):
    P = (STRING,)

    def execute(self, vm):
        vm.store_memory(vm.dc, len(self.args[0]))
        vm.dc += 1
        for c in self.args[0]:
            vm.store_memory(vm.dc, ord(c))
            vm.dc += 1

    def assemble(self):
        s = self.args[0]
        length_bytes = [len(s) & 0xFF00, len(s) & 0xFF]
        data_bytes = []
        for c in s:
            data_bytes.append(0)
            data_bytes.append(ord(c))
        return bytes(length_bytes + data_bytes)


class CONSTANT(DataOperation):
    P = (LABEL_TYPE, I16)

    def convert(self):
        return []


class LABEL(AbstractOperation):
    P = (LABEL_TYPE,)

    def convert(self):
        return []


class DLABEL(DataOperation):
    P = (LABEL_TYPE,)

    def convert(self):
        return []


class PRINT_REG(DebuggingOperation):
    P = (REGISTER,)

    def execute(self, vm):
        v = vm.load_register(self.args[0])
        print("R{} = {}".format(self.args[0], format_int(v)))
        vm.pc += 1


class PRINT(DebuggingOperation):
    P = (STRING,)

    def execute(self, vm):
        print(self.args[0], end="")
        vm.pc += 1


class PRINTLN(DebuggingOperation):
    P = (STRING,)

    def execute(self, vm):
        print(self.args[0])
        vm.pc += 1


class __EVAL(DebuggingOperation):
    P = (STRING,)

    def execute(self, vm):
        try:
            eval(self.args[0], {}, {"stdlib": stdlib, "vm": vm})
        except Exception as e:
            print_error(vm.settings, "Python exception: " + str(e), loc=vm.location)
            sys.exit(3)
        vm.pc += 1


def disassemble(v: int) -> AbstractOperation:
    for cls in name_to_class.values():
        if cls.BITV != "":
            m = match_bitvector(cls.BITV, v)
            if m is not False and isinstance(m, list):
                return cls.disassemble(*m)

    raise HERAError("bit pattern does not correspond to HERA instruction")


def match_bitvector(pattern: str, v: int) -> Union[List, bool]:
    """
    Try to match the 16-bit integer `v` against `pattern`. Return a list of extracted
    arguments if `v` matches, or False otherwise.

    `pattern` should be a string of sixteen characters, which may be the digits '0' or
    '1' or Latin letters. The digits are matched against the literal digits in `v`;
    letters are used to extract values. Uppercase letters match integer values and
    lowercase letters match registers. For example, "0000aaaaBBBB1111" would match,
    e.g., 0000 1001 0110 11111, returning

        [Token(Token.INT, 0b1001), Token(Token.REGISTER, 0b0110)]
    """
    s = bin(v)[2:].rjust(16, "0")
    args = []  # type: List[Token]

    pattern = pattern.replace(" ", "")

    # Initialize argument list with the correct types (integers or registers).
    for pattern_bit in pattern:
        if pattern_bit.isalpha():
            index = ord(pattern_bit.lower()) - ord("a")
            while index >= len(args):
                args.append(Token(Token.INT, 0))

            if pattern_bit.isupper():
                args[index] = Token(Token.REGISTER, 0)
            else:
                args[index] = Token(Token.INT, 0)

    # Fill in the argument list with the actual values from the binary data.
    for pattern_bit, real_bit in zip(pattern, s):
        if pattern_bit == "0":
            if real_bit != "0":
                return False
        elif pattern_bit == "1":
            if real_bit != "1":
                return False
        else:
            index = ord(pattern_bit.lower()) - ord("a")
            args[index].value <<= 1
            if real_bit == "1":
                args[index].value += 1

    return args


def substitute_bitvector(pattern: str, args: List[int]) -> bytes:
    # Make a copy of the arguments as we will be modifying it.
    args = args[:]

    # Remove spaces from the pattern string.
    pattern = pattern.replace(" ", "")

    # Substituting the low byte MUST happen before the high byte, because it pulls the
    # low bits from each argument.
    low_pattern = pattern[8:16]
    low_byte = substitute_half_a_bitvector(low_pattern, args)

    high_pattern = pattern[0:8]
    high_byte = substitute_half_a_bitvector(high_pattern, args)

    return bytes([high_byte, low_byte])


def substitute_half_a_bitvector(pattern: str, args: List[int]) -> int:
    ret = 0
    for shift, pattern_bit in enumerate(reversed(pattern)):
        if pattern_bit == "0":
            real_bit = 0
        elif pattern_bit == "1":
            real_bit = 1
        else:
            index = ord(pattern_bit.lower()) - ord("a")
            real_bit = args[index] & 1
            args[index] >>= 1

        ret += real_bit << shift

    return ret


def check_arglist(argtypes, args, symbol_table):
    messages = Messages()
    for expected, got in zip(argtypes, args):
        if expected == REGISTER:
            err = check_register(got)
        elif expected == REGISTER_OR_LABEL:
            err = check_register_or_label(got, symbol_table)
        elif expected == LABEL_TYPE:
            err = check_label(got)
        elif expected == STRING:
            err = check_string(got)
        elif expected == I16_OR_LABEL:
            err = check_in_range(
                got, symbol_table, lo=-2 ** 15, hi=2 ** 16, labels=True
            )
        elif expected == I8_OR_LABEL:
            err = check_in_range(got, symbol_table, lo=-2 ** 7, hi=2 ** 8, labels=True)
        elif isinstance(expected, range):
            err = check_in_range(got, symbol_table, lo=expected.start, hi=expected.stop)
        else:
            raise RuntimeError("unknown parameter type {!r}".format(expected))

        if err is not None:
            messages.err(err, got)
    return messages


def check_register(arg) -> Optional[str]:
    if arg.type == Token.REGISTER:
        return None
    else:
        if isinstance(arg.value, str) and arg.value.lower() == "pc":
            return "program counter cannot be accessed or changed directly"
        else:
            return "expected register"


def check_register_or_label(arg, symbol_table: Dict[str, int]) -> Optional[str]:
    if arg.type == Token.REGISTER:
        return None
    elif arg.type == Token.SYMBOL:
        try:
            val = symbol_table[arg.value]
        except KeyError:
            return "undefined symbol"
        else:
            if isinstance(val, Constant):
                return "constant cannot be used as label"
            elif isinstance(val, DataLabel):
                return "data label cannot be used as branch label"
            else:
                return None
    else:
        return "expected register or label"


def check_label(arg) -> Optional[str]:
    if arg.type == Token.SYMBOL:
        return None
    else:
        return "expected label"


def check_string(arg):
    if not isinstance(arg, Token) or arg.type != Token.STRING:
        return "expected string literal"
    else:
        return None


def check_in_range(arg, symbol_table, *, lo, hi, labels=False):
    if arg.type == Token.SYMBOL:
        try:
            arg = Token.Int(symbol_table[arg.value])
        except KeyError:
            return "undefined constant"
        else:
            if not labels and not isinstance(arg.value, Constant):
                return "cannot use label as constant"
            elif labels and isinstance(arg.value, Label):
                return None

    if arg.type != Token.INT:
        return "expected integer"

    if arg.value < lo or arg.value >= hi:
        return "integer must be in range [{}, {})".format(lo, hi)
    else:
        return None


name_to_class = {
    "ADD": ADD,
    "AND": AND,
    "ASL": ASL,
    "ASR": ASR,
    "BC": BC,
    "BCR": BCR,
    "BG": BG,
    "BGR": BGR,
    "BGE": BGE,
    "BGER": BGER,
    "BL": BL,
    "BLR": BLR,
    "BLE": BLE,
    "BLER": BLER,
    "BNC": BNC,
    "BNCR": BNCR,
    "BNS": BNS,
    "BNSR": BNSR,
    "BNV": BNV,
    "BNVR": BNVR,
    "BNZ": BNZ,
    "BNZR": BNZR,
    "BR": BR,
    "BRR": BRR,
    "BS": BS,
    "BSR": BSR,
    "BUG": BUG,
    "BUGR": BUGR,
    "BULE": BULE,
    "BULER": BULER,
    "BV": BV,
    "BVR": BVR,
    "BZ": BZ,
    "BZR": BZR,
    "CALL": CALL,
    "CBON": CBON,
    "CCBOFF": CCBOFF,
    "CMP": CMP,
    "COFF": COFF,
    "CON": CON,
    "CONSTANT": CONSTANT,
    "DEC": DEC,
    "DLABEL": DLABEL,
    "DSKIP": DSKIP,
    "FLAGS": FLAGS,
    "FOFF": FOFF,
    "FON": FON,
    "FSET4": FSET4,
    "FSET5": FSET5,
    "HALT": HALT,
    "INC": INC,
    "INTEGER": INTEGER,
    "LABEL": LABEL,
    "LOAD": LOAD,
    "LP_STRING": LP_STRING,
    "LSL": LSL,
    "LSL8": LSL8,
    "LSR": LSR,
    "LSR8": LSR8,
    "MOVE": MOVE,
    "MUL": MUL,
    "NEG": NEG,
    "NOP": NOP,
    "NOT": NOT,
    "OPCODE": OPCODE,
    "OR": OR,
    "print": PRINT,
    "println": PRINTLN,
    "print_reg": PRINT_REG,
    "RETURN": RETURN,
    "RSTRF": RSTRF,
    "RTI": RTI,
    "SAVEF": SAVEF,
    "SET": SET,
    "SETHI": SETHI,
    "SETLO": SETLO,
    "SETRF": SETRF,
    "STORE": STORE,
    "SUB": SUB,
    "SWI": SWI,
    "TIGER_STRING": LP_STRING,
    "XOR": XOR,
    "__eval": __EVAL,
}



import pytest

from hera.main import main
from .utils import execute_program_helper


def test_basic_include():
    program = """\
#include "test/assets/include/lib/add.hera"

SET(R1, 20)
SET(R2, 22)
CALL(R12, add)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 20
    assert vm.registers[2] == 22
    assert vm.registers[3] == 42

    for r in vm.registers[4:11]:
        assert r == 0

    assert not vm.flag_sign
    assert not vm.flag_zero
    assert not vm.flag_overflow
    assert not vm.flag_carry
    assert not vm.flag_carry_block


def test_multiple_includes():
    program = """\
#include "test/assets/include/lib/r1_to_42.hera"
#include "test/assets/include/lib/r2_to_42.hera"

SET(R3, 42)
    """
    vm = execute_program_helper(program)

    assert vm.registers[1] == 42
    assert vm.registers[2] == 42
    assert vm.registers[3] == 42


def test_recursive_program(capsys):
    with pytest.raises(SystemExit):
        main(["test/assets/include/recursive.hera"])

    captured = capsys.readouterr()
    assert "recursive include" in captured.err
    assert '#include "recursive.hera"' in captured.err
    assert "line 1 col 10 of test/assets/include/recursive.hera" in captured.err


def test_mutually_recursive_programs(capsys):
    with pytest.raises(SystemExit):
        main(["test/assets/include/mutually_recursive1.hera"])

    captured = capsys.readouterr()
    assert "recursive include" in captured.err
    assert '#include "mutually_recursive1.hera"' in captured.err
    assert (
        "line 1 col 10 of test/assets/include/mutually_recursive2.hera" in captured.err
    )


def test_nonexistent_path_program(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper('#include "test/assets/include/whatever.hera"')

    captured = capsys.readouterr()
    assert 'file "test/assets/include/whatever.hera" does not exist' in captured.err


@pytest.mark.skip("not yet")
def test_include_stdin_program(capsys):
    program = '#include "-"\n#include "<stdin>"'
    with pytest.raises(SystemExit):
        execute_program_helper(program)

    captured = capsys.readouterr()
    assert 'file "-" does not exist' in captured.err
    # We want this error and not a recursive include error, i.e. we need to distinguish
    # between actual standard input and a file called "<stdin>".
    assert 'file "<stdin>" does not exist' in captured.err



"""
Utilities for the hera-py system.

Author:  Ian Fisher (iafisher@protonmail.com)
Version: February 2019
"""
import sys

from .data import HERAError, Location, Messages, Settings, Token


def to_u16(n: int) -> int:
    """
    Reinterpret the signed integer `n` as a 16-bit unsigned integer.

    If `n` is too large for 16 bits, a HERAError is raised.
    """
    # Note that we allow positive values up to 2**16, but negative values only
    # down to -2**15.
    if n >= 2 ** 16 or n < -2 ** 15:
        raise HERAError("signed integer too large for 16 bits")

    if n < 0:
        return 2 ** 16 + n
    else:
        return n


def from_u16(n: int) -> int:
    """Reinterpret the unsigned 16-bit integer `n` as a signed integer."""
    if n >= 2 ** 15:
        return -(2 ** 16 - n)
    else:
        return n


def to_u32(n: int) -> int:
    """
    Reinterpret the signed integer `n` as an unsigned 32-bit integer.

    If `n` is too large for 32 bits, a HERAError is raised.
    """
    # Note that we allow positive values up to 2**32, but negative values only
    # down to -2**31.
    if n >= 2 ** 32 or n < -2 ** 31:
        raise HERAError("signed integer too large for 16 bits")

    if n < 0:
        return 2 ** 32 + n
    else:
        return n


NAMED_REGISTERS = {"rt": 11, "fp": 14, "sp": 15, "pc_ret": 13, "fp_alt": 12}


def register_to_index(rname: str) -> int:
    """Return the index of the register with the given name in the register array."""
    original = rname
    rname = rname.lower()
    if rname in NAMED_REGISTERS:
        return NAMED_REGISTERS[rname]
    elif rname.startswith("r"):
        v = int(rname[1:])
        if 0 <= v < 16:
            return v
    raise HERAError("{} is not a valid register".format(original))


def format_int(v: int, *, spec="xdsc") -> str:
    """
    Return a string of the form "... = ... = ..." where each ellipsis stands for a
    formatted integer determined by a character in the `spec` parameter. The following
    formats are supported: d for decimal, x for hexadecimal, o for octal, b for binary,
    c for character literal, and s for signed integer. The latter two formats only
    generate output when applicable, e.g. for integers that actually represent printable
    characters and signed integers, respectively. Output can be forced for either of
    these formats by capitalizing the letter.
    """
    ret = []
    for c in spec:
        if c == "d":
            ret.append(str(v))
        elif c == "x":
            ret.append("0x{:0>4x}".format(v))
        elif c == "o":
            ret.append("0o{:0>8o}".format(v))
        elif c == "b":
            ret.append("0b{:0>16b}".format(v))
        elif c == "c":
            if v < 128 and chr(v).isprintable():
                ret.append(repr(chr(v)))
        elif c == "C":
            if v < 128:
                ret.append(repr(chr(v)))
            else:
                ret.append("not an ASCII character")
        elif c == "s":
            if v & 0x8000:
                ret.append(str(from_u16(v)))
        elif c == "S":
            if v & 0x8000:
                ret.append(str(from_u16(v)))
            else:
                ret.append("not a signed integer")
        else:
            raise RuntimeError("unknown format specifier: " + c)
    return " = ".join(ret)


def print_warning(settings: Settings, msg: str, *, loc=None) -> None:
    if settings.color:
        msg = ANSI_MAGENTA_BOLD + "Warning" + ANSI_RESET + ": " + msg
    else:
        msg = "Warning: " + msg
    print_message(msg, loc=loc)


def print_error(settings: Settings, msg: str, *, loc=None) -> None:
    if settings.color:
        msg = ANSI_RED_BOLD + "Error" + ANSI_RESET + ": " + msg
    else:
        msg = "Error: " + msg
    print_message(msg, loc=loc)


def print_message(msg: str, *, loc=None) -> None:
    """
    Print a message to stderr. If `loc` is provided as either a Location object, or a
    Token object with a `location` field, then the line of code that the location
    indicates will be printed with the message.
    """
    if isinstance(loc, Token):
        loc = loc.location

    if isinstance(loc, Location):
        linetext = loc.file_lines[loc.line - 1]
        caret = align_caret(linetext, loc.column) + "^"
        msg += ", line {} col {} of {}\n\n  {}\n  {}\n".format(
            loc.line, loc.column, loc.path, linetext, caret
        )

    sys.stderr.write(msg + "\n")


def align_caret(line: str, col: int) -> str:
    """
    Return the whitespace necessary to align a caret to underline the desired column in
    the line of text. Mainly this means handling tabs.
    """
    return "".join("\t" if c == "\t" else " " for c in line[: col - 1])


def read_file(path: str) -> str:
    """Read a file and return its contents."""
    try:
        with open(path, encoding="ascii") as f:
            return f.read()
    except FileNotFoundError:
        raise HERAError('file "{}" does not exist'.format(path))
    except PermissionError:
        raise HERAError('permission denied to open file "{}"'.format(path))
    except OSError:
        raise HERAError('could not open file "{}"'.format(path))
    except UnicodeDecodeError:
        raise HERAError('non-ASCII byte in file "{}"'.format(path))


def read_file_or_stdin(path: str, settings) -> str:
    if path == "-":
        try:
            text = sys.stdin.read()
        except (IOError, KeyboardInterrupt):
            # Print to stderr when in interpreter or debug mode, because the output of
            # the HERA program goes to stdout.
            if settings.mode in ("", "debug"):
                print(file=sys.stderr)
            else:
                print()
            sys.exit(3)
        else:
            # So that the program and its output are visually separate.
            if settings.mode in ("", "debug"):
                print(file=sys.stderr)
            else:
                print()

        try:
            text.encode("ascii")
        except UnicodeEncodeError:
            handle_messages(settings, Messages("non-ASCII byte in file."))
    else:
        try:
            text = read_file(path)
        except HERAError as e:
            handle_messages(settings, Messages(str(e) + "."))

    return text


def pad(s: str, n: int) -> str:
    return (" " * (n - len(s))) + s


def handle_messages(settings: Settings, messages: Messages) -> None:
    """
    Print to standard error for any warnings or errors recorded in `messages`. If any
    errors were recorded, exit the program.
    """
    for msg, loc in messages.warnings:
        print_warning(settings, msg, loc=loc)

    settings.warning_count += len(messages.warnings)
    messages.warnings.clear()

    for msg, loc in messages.errors:
        print_error(settings, msg, loc=loc)

    if messages.errors:
        sys.exit(3)


# ANSI color codes (https://stackoverflow.com/questions/4842424/)
# When the --no-color flag is specified, these constants are set to the empty string, so
# you can use them unconditionally in your code without worrying about --no-color.


def make_ansi(*params) -> str:
    return "\033[" + ";".join(map(str, params)) + "m"


ANSI_RED_BOLD = make_ansi(31, 1)
ANSI_MAGENTA_BOLD = make_ansi(35, 1)
ANSI_RESET = make_ansi(0)
